from autosar.element import Element

class SwcImplementation(Element):
    """ Swc implementation """
    def __init__(self,name, behaviorRef, parent=None):
        super().__init__(name, parent)
        self.behaviorRef = behaviorRef
        self.codeDescriptors = None
        self.programmingLanguage = None
        self.resourceConsumption = None
        self.swVersion = None
        self.useCodeGenerator = None
        self.vendorId = None

class SwcImplementationCodeDescriptor(Element):
    """ Swc implementation code """
    def __init__(self,name, parent=None):
        super().__init__(name, parent)
        # Autosar 4
        self.artifactDescriptors = None

        # Autosar 3
        self.type = None

class EngineeringObject(object):
    """ EngineeringObject """
    def __init__(self, parent=None):
        self.parent = parent
        self.shortLabel = None
        self.category = None
        self.revisionLabels = None
        self.domain = None

class ResourceConsumption(Element):
    """ Swc implementation ResourceConsumption """
    def __init__(self,name, parent=None):
        super().__init__(name, parent)
        self.memorySections = None

class MemorySection(Element):
    """ Swc implementation MemorySection """
    def __init__(self,name, parent=None):
        super().__init__(name, parent)
        self.aligment = None
        self.memClassSymbol = None
        self.options = None
        self.size = None
        self.swAddrmethodRef = None
        self.symbol = None



"""
Implements the class SwcImplementationParser
"""
import sys
from autosar.base import splitRef, hasAdminData, parseAdminDataNode
import autosar.component
from autosar.parser.parser_base import ElementParser

class SwcImplementationParser(ElementParser):
    """
    ComponentType parser
    """

    def __init__(self,version=3.0):
        super().__init__(version)
        self.classTag = 'SWC-IMPLEMENTATION'

    def getSupportedTags(self):
        return [self.classTag]

    def parseElement(self, xmlElement, parent = None):
        """
        parser for the class.
        """
        assert (xmlElement.tag=='SWC-IMPLEMENTATION')
        ws = parent.rootWS()
        assert(ws is not None)
        name = self.parseTextNode(xmlElement.find('SHORT-NAME'))
        behaviorRef = self.parseTextNode(xmlElement.find('BEHAVIOR-REF'))
        implementation = autosar.component.SwcImplementation(name, behaviorRef, parent=parent)
        self.push()
        for xmlElem in xmlElement.findall('./*'):
            if xmlElem.tag   == 'SHORT-NAME':
                continue
            elif xmlElem.tag == 'BEHAVIOR-REF':
                continue
            elif xmlElem.tag == 'BUILD-ACTION-MANIFESTS':
                #TODO: Implement later
                continue
            elif xmlElem.tag == 'CODE-DESCRIPTORS':
                # Create the list to indicate that the base element exists.
                implementation.codeDescriptors = []
                for codeElem in xmlElem.findall('./*'):
                    if codeElem.tag  == 'CODE':
                        implementation.codeDescriptors.append(self.parseCodeDescriptor(codeElem, parent=implementation))
                    else:
                        raise NotImplementedError(codeElem.tag)
            elif xmlElem.tag == 'COMPILERS':
                #TODO: Implement later
                continue
            elif xmlElem.tag == 'GENERATED-ARTIFACTS':
                #TODO: Implement later
                continue
            elif xmlElem.tag == 'HW-ELEMENT-REFS':
                #TODO: Implement later
                continue
            elif xmlElem.tag == 'LINKERS':
                #TODO: Implement later
                continue
            elif xmlElem.tag == 'MC-SUPPORT':
                #TODO: Implement later
                continue
            elif xmlElem.tag == 'PROGRAMMING-LANGUAGE':
                implementation.programmingLanguage = self.parseTextNode(xmlElem)
                continue
            elif xmlElem.tag == 'REQUIRED-ARTIFACTS':
                #TODO: Implement later
                continue
            elif xmlElem.tag == 'REQUIRED-GENERATOR-TOOLS':
                #TODO: Implement later
                continue
            elif xmlElem.tag == 'RESOURCE-CONSUMPTION':
                implementation.resourceConsumption = self.parseResourceConsumption(xmlElem, parent=implementation)
                continue
            elif xmlElem.tag == 'SW-VERSION':
                implementation.swVersion = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'SWC-BSW-MAPPING-REF':
                #TODO: Implement later
                continue
            elif xmlElem.tag == 'USED-CODE-GENERATOR':
                implementation.useCodeGenerator = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'VENDOR-ID':
                implementation.vendorId = self.parseIntNode(xmlElem)
            else:
                self.defaultHandler(xmlElem)

        self.pop(implementation)

        # Find the SWC and connect the swc to the implementation.
        behavior = ws.find(behaviorRef)
        if behavior is not None:
            swc = ws.find(behavior.componentRef)
            if swc is not None:
                swc.implementation = implementation

        return implementation

    def parseCodeDescriptor(self, xmlElement, parent = None):
        """
        Parser for implementation code descriptor.
        """
        assert (xmlElement.tag=='CODE')
        name = self.parseTextNode(xmlElement.find('SHORT-NAME'))
        code = autosar.component.SwcImplementationCodeDescriptor(name, parent=parent)
        self.push()
        for xmlElem in xmlElement.findall('./*'):
            if xmlElem.tag   == 'SHORT-NAME':
                continue
            elif xmlElem.tag == 'ARTIFACT-DESCRIPTORS':
                code.artifactDescriptors = []
                for descElem in xmlElem.findall('./*'):
                    if descElem.tag  == 'AUTOSAR-ENGINEERING-OBJECT':
                        engineeringObject = autosar.component.EngineeringObject(code)
                        for elem in descElem.findall('./*'):
                            if elem.tag  == 'SHORT-LABEL':
                                engineeringObject.shortLabel = self.parseTextNode(elem)
                            elif elem.tag  == 'CATEGORY':
                                engineeringObject.category = self.parseTextNode(elem)
                            elif elem.tag == 'REVISION-LABELS':
                                engineeringObject.revisionLabels = []
                                for labelElem in elem.findall('./*'):
                                    if labelElem.tag  == 'REVISION-LABEL':
                                        engineeringObject.revisionLabels.append(self.parseTextNode(labelElem))
                            elif elem.tag  == 'DOMAIN':
                                engineeringObject.domain = self.parseTextNode(elem)
                        code.artifactDescriptors.append(engineeringObject)
            elif xmlElem.tag == 'CALLBACK-HEADER-REFS':
                #TODO: Implement later
                continue
            elif xmlElem.tag == 'TYPE':
                # Only valid in Autosar 3.
                code.type = self.parseTextNode(xmlElem)
            else:
                self.defaultHandler(xmlElem)

        self.pop(code)
        return code

    def parseResourceConsumption(self, xmlElement, parent = None):
        """
        Parser for implementation resource consumption.
        """
        assert (xmlElement.tag=='RESOURCE-CONSUMPTION')
        name = self.parseTextNode(xmlElement.find('SHORT-NAME'))
        res = autosar.component.ResourceConsumption(name, parent=parent)
        self.push()
        for xmlElem in xmlElement.findall('./*'):
            if xmlElem.tag   == 'SHORT-NAME':
                continue
            elif xmlElem.tag == 'EXECUTION-TIMES':
                #TODO: Implement later
                continue
            elif xmlElem.tag == 'HEAP-USAGES':
                #TODO: Implement later
                continue
            elif xmlElem.tag == 'MEMORY-SECTIONS':
                res.memorySections = []
                for xmlSection in xmlElem.findall('./*'):
                    res.memorySections.append(self.parseMemorySection(xmlSection, res))
            elif xmlElem.tag == 'SECTION-NAME-PREFIXS':
                #TODO: Implement later
                continue
            elif xmlElem.tag == 'STACK-USAGES':
                #TODO: Implement later
                continue
            else:
                self.defaultHandler(xmlElem)

        self.pop(res)
        return res

    def parseMemorySection(self, xmlElement, parent = None):
        """
        Parser for memory section.
        """
        assert (xmlElement.tag=='MEMORY-SECTION')
        name = self.parseTextNode(xmlElement.find('SHORT-NAME'))
        section = autosar.component.MemorySection(name, parent=parent)
        self.push()
        for xmlElem in xmlElement.findall('./*'):
            if xmlElem.tag   == 'SHORT-NAME':
                continue
            elif xmlElem.tag == 'ALIGNMENT':
                section.aligment = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'EXECUTABLE-ENTITY-REFS':
                #TODO: Implement later
                continue
            elif xmlElem.tag == 'MEM-CLASS-SYMBOL':
                section.memClassSymbol = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'OPTIONS':
                section.options = []
                for xmlOption in xmlElem.findall('./*'):
                    section.options.append(self.parseTextNode(xmlOption))
            elif xmlElem.tag == 'PREFIX-REF':
                #TODO: Implement later
                continue
            elif xmlElem.tag == 'SIZE':
                section.size = self.parseIntNode(xmlElem)
            elif xmlElem.tag == 'SW-ADDRMETHOD-REF':
                section.swAddrmethodRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'SYMBOL':
                section.symbol = self.parseTextNode(xmlElem)
            else:
                self.defaultHandler(xmlElem)

        self.pop(section)
        return section



import autosar.behavior
from autosar.parser.parser_base import ElementParser
from autosar.parser.constant_parser import ConstantParser

class BehaviorParser(ElementParser):
    def __init__(self,version=3.0):
        super().__init__(version)
        self.constantParser = ConstantParser(version)

    def getSupportedTags(self):
        if (self.version >=3.0) and (self.version < 4.0):
            return ['INTERNAL-BEHAVIOR']
        elif self.version >= 4.0:
            return ['SWC-INTERNAL-BEHAVIOR']
        else:
            return []

    def parseElement(self, xmlElement, parent = None):
        if (self.version >=3.0) and (self.version < 4.0) and xmlElement.tag == 'INTERNAL-BEHAVIOR':
            return self.parseInternalBehavior(xmlElement, parent)
        elif (self.version >= 4.0) and (xmlElement.tag == 'SWC-INTERNAL-BEHAVIOR'):
            return self.parseSWCInternalBehavior(xmlElement, parent)
        else:
            return None

    def parseInternalBehavior(self,xmlRoot,parent):
        """AUTOSAR 3 Internal Behavior"""
        assert(xmlRoot.tag == 'INTERNAL-BEHAVIOR')
        name = self.parseTextNode(xmlRoot.find('SHORT-NAME'))
        componentRef = self.parseTextNode(xmlRoot.find('COMPONENT-REF'))
        multipleInstance = False
        xmlSupportMultipleInst = xmlRoot.find('SUPPORTS-MULTIPLE-INSTANTIATION')
        if (xmlSupportMultipleInst is not None) and (xmlSupportMultipleInst.text == 'true'):
            multipleInstance = True
        ws = parent.rootWS()
        assert(ws is not None)
        if (name is not None) and (componentRef is not None):
            internalBehavior = autosar.behavior.InternalBehavior(name, componentRef, multipleInstance, parent)
            swc = ws.find(componentRef)
            if swc is not None:
                swc.behavior=internalBehavior
            for xmlNode in xmlRoot.findall('./*'):
                if (xmlNode.tag == 'SHORT-NAME') or (xmlNode.tag == 'COMPONENT-REF') or (xmlNode.tag == 'SUPPORTS-MULTIPLE-INSTANTIATION'):
                    continue
                if xmlNode.tag == 'EVENTS':
                    for xmlEvent in xmlNode.findall('./*'):
                        event = None
                        if xmlEvent.tag == 'MODE-SWITCH-EVENT':
                            event = self.parseModeSwitchEvent(xmlEvent,internalBehavior)
                        elif xmlEvent.tag == 'TIMING-EVENT':
                            event = self.parseTimingEvent(xmlEvent,internalBehavior)
                        elif xmlEvent.tag == 'DATA-RECEIVED-EVENT':
                            event = self.parseDataReceivedEvent(xmlEvent,internalBehavior)
                        elif xmlEvent.tag == 'OPERATION-INVOKED-EVENT':
                            event = self.parseOperationInvokedEvent(xmlEvent,internalBehavior)
                        else:
                            raise NotImplementedError(xmlEvent.tag)
                        if event is not None:
                            internalBehavior.events.append(event)
                        else:
                            raise ValueError('event')
                elif xmlNode.tag == 'PORT-API-OPTIONS':
                    for xmlOption in xmlNode.findall('./PORT-API-OPTION'):
                        portAPIOption = autosar.behavior.PortAPIOption(self.parseTextNode(xmlOption.find('PORT-REF')),
                                                                       self.parseBooleanNode(xmlOption.find('ENABLE-TAKE-ADDRESS')),
                                                                       self.parseBooleanNode(xmlOption.find('INDIRECT-API')))
                        if portAPIOption is not None: internalBehavior.portAPIOptions.append(portAPIOption)
                elif xmlNode.tag == 'RUNNABLES':
                    for xmRunnable in xmlNode.findall('./RUNNABLE-ENTITY'):
                        runnableEntity = self.parseRunnableEntity(xmRunnable, internalBehavior)
                        if runnableEntity is not None:
                            internalBehavior.runnables.append(runnableEntity)
                elif xmlNode.tag == 'PER-INSTANCE-MEMORYS':
                    for xmlElem in xmlNode.findall('./PER-INSTANCE-MEMORY'):
                        perInstanceMemory = autosar.behavior.PerInstanceMemory(self.parseTextNode(xmlElem.find('SHORT-NAME')),self.parseTextNode(xmlElem.find('TYPE-DEFINITION')), internalBehavior)
                        if perInstanceMemory is not None: internalBehavior.perInstanceMemories.append(perInstanceMemory)
                elif xmlNode.tag == 'SERVICE-NEEDSS':
                    for xmlElem in xmlNode.findall('./*'):
                        if xmlElem.tag=='SWC-NV-BLOCK-NEEDS':
                            swcNvBlockNeeds=self.parseSwcNvBlockNeeds(xmlElem)
                            if swcNvBlockNeeds is not None: internalBehavior.swcNvBlockNeeds.append(swcNvBlockNeeds)
                        else:
                            raise NotImplementedError(xmlElem.tag)
                elif xmlNode.tag == 'SHARED-CALPRMS':
                    for xmlElem in xmlNode.findall('./*'):
                        if xmlElem.tag=='CALPRM-ELEMENT-PROTOTYPE':
                            calPrmElemPrototype=self.parseCalPrmElemPrototype(xmlElem, internalBehavior)
                            assert(calPrmElemPrototype is not None)
                            internalBehavior.sharedCalParams.append(calPrmElemPrototype)
                        else:
                            raise NotImplementedError(xmlElem.tag)
                elif xmlNode.tag == 'EXCLUSIVE-AREAS':
                    for xmlElem in xmlNode.findall('./*'):
                        if xmlElem.tag=='EXCLUSIVE-AREA':
                            exclusiveArea=autosar.behavior.ExclusiveArea(self.parseTextNode(xmlElem.find('SHORT-NAME')), internalBehavior)
                            internalBehavior.exclusiveAreas.append(exclusiveArea)
                        else:
                            raise NotImplementedError(xmlElem.tag)
                else:
                    raise NotImplementedError(xmlNode.tag)
            return internalBehavior

    def parseSWCInternalBehavior(self, xmlRoot, parent):
        """AUTOSAR 4 internal behavior"""
        assert(xmlRoot.tag == 'SWC-INTERNAL-BEHAVIOR')
        name = self.parseTextNode(xmlRoot.find('SHORT-NAME'))
        multipleInstance = False
        xmlSupportMultipleInst = xmlRoot.find('SUPPORTS-MULTIPLE-INSTANTIATION')
        if xmlSupportMultipleInst is not None:
            multipleInstance = self.parseBooleanNode(xmlSupportMultipleInst)
            assert(multipleInstance is not None)
        ws = parent.rootWS()
        assert(ws is not None)
        if (name is not None):
            handledXML = ['SHORT-NAME', 'SUPPORTS-MULTIPLE-INSTANTIATION']
            internalBehavior = autosar.behavior.SwcInternalBehavior(name, parent.ref, multipleInstance, parent)
            for xmlElem in xmlRoot.findall('./*'):
                if xmlElem.tag in handledXML:
                    pass
                elif xmlElem.tag == 'DATA-TYPE-MAPPING-REFS':
                    for xmlChild in xmlElem.findall('./*'):
                        if xmlChild.tag == 'DATA-TYPE-MAPPING-REF':
                            tmp = self.parseTextNode(xmlChild)
                            assert(tmp is not None)
                            internalBehavior.appendDataTypeMappingRef(tmp)
                elif xmlElem.tag == 'EVENTS':
                    for xmlEvent in xmlElem.findall('./*'):
                        event = None
                        if xmlEvent.tag == 'INIT-EVENT':
                            event = self.parseInitEvent(xmlEvent, internalBehavior)
                        elif xmlEvent.tag == 'SWC-MODE-SWITCH-EVENT':
                            event = self.parseModeSwitchEvent(xmlEvent, internalBehavior)
                        elif xmlEvent.tag == 'TIMING-EVENT':
                            event = self.parseTimingEvent(xmlEvent, internalBehavior)
                        elif xmlEvent.tag == 'DATA-RECEIVED-EVENT':
                            event = self.parseDataReceivedEvent(xmlEvent, internalBehavior)
                        elif xmlEvent.tag == 'OPERATION-INVOKED-EVENT':
                            event = self.parseOperationInvokedEvent(xmlEvent, internalBehavior)
                        elif xmlEvent.tag == 'MODE-SWITCHED-ACK-EVENT':
                            event = self.parseModeSwitchedAckEvent(xmlEvent, internalBehavior)
                        elif xmlEvent.tag == 'DATA-RECEIVE-ERROR-EVENT':
                            #TODO: Implement later
                            pass
                        else:
                            raise NotImplementedError(xmlEvent.tag)
                        if event is not None:
                            internalBehavior.events.append(event)
                elif xmlElem.tag == 'PORT-API-OPTIONS':
                    for xmlOption in xmlElem.findall('./PORT-API-OPTION'):
                        portRef, takeAddress, indirectAPI, errorHandling = self.parsePortAPIOptions(xmlOption)
                        portAPIOption = autosar.behavior.PortAPIOption(portRef, takeAddress, indirectAPI, errorHandling)
                        if portAPIOption is not None: internalBehavior.portAPIOptions.append(portAPIOption)
                elif xmlElem.tag == 'RUNNABLES':
                    for xmRunnable in xmlElem.findall('./RUNNABLE-ENTITY'):
                        runnableEntity = self.parseRunnableEntity(xmRunnable, internalBehavior)
                        if runnableEntity is not None:
                            internalBehavior.runnables.append(runnableEntity)
                elif xmlElem.tag == 'AR-TYPED-PER-INSTANCE-MEMORYS':
                    for xmlChild in xmlElem.findall('./*'):
                        if xmlChild.tag == 'VARIABLE-DATA-PROTOTYPE':
                            dataElement = self.parseVariableDataPrototype(xmlChild, internalBehavior)
                            internalBehavior.perInstanceMemories.append(dataElement)
                        else:
                            raise NotImplementedError(xmlChild.tag)
                elif xmlElem.tag == 'SERVICE-DEPENDENCYS':
                    for xmlChildElem in xmlElem.findall('./*'):
                        if xmlChildElem.tag == 'SWC-SERVICE-DEPENDENCY':
                            swcServiceDependency = self.parseSwcServiceDependency(xmlChildElem, internalBehavior)
                            internalBehavior.serviceDependencies.append(swcServiceDependency)
                        else:
                            raise NotImplementedError(xmlChildElem.tag)
                elif xmlElem.tag == 'SHARED-PARAMETERS':
                    for xmlChildElem in xmlElem.findall('./*'):
                        if xmlChildElem.tag == 'PARAMETER-DATA-PROTOTYPE':
                            tmp = self.parseParameterDataPrototype(xmlChildElem, internalBehavior)
                            if tmp is not None:
                                internalBehavior.parameterDataPrototype.append(tmp)
                        else:
                            raise NotImplementedError(xmlChildElem.tag)
                elif xmlElem.tag == 'EXCLUSIVE-AREAS':
                    for xmlChild in xmlElem.findall('./*'):
                        if xmlChild.tag=='EXCLUSIVE-AREA':
                            exclusiveArea=autosar.behavior.ExclusiveArea(self.parseTextNode(xmlChild.find('SHORT-NAME')), internalBehavior)
                            internalBehavior.exclusiveAreas.append(exclusiveArea)
                        else:
                            raise NotImplementedError(xmlChild.tag)
                elif xmlElem.tag == 'PER-INSTANCE-PARAMETERS':
                    pass #implement later
                elif xmlElem.tag == 'EXPLICIT-INTER-RUNNABLE-VARIABLES':
                    pass #implement later
                elif xmlElem.tag == 'HANDLE-TERMINATION-AND-RESTART':
                    pass #implement later
                elif xmlElem.tag == 'STATIC-MEMORYS':
                    pass #implement later
                else:
                    raise NotImplementedError(xmlElem.tag)
            return internalBehavior

    def parsePortAPIOptions(self, xmlElem):
        portRef = self.parseTextNode(xmlElem.find('PORT-REF'))
        takeAddress = self.parseBooleanNode(xmlElem.find('ENABLE-TAKE-ADDRESS'))
        indirectAPI = self.parseBooleanNode(xmlElem.find('INDIRECT-API'))
        errorHandling = self.parseTextNode(xmlElem.find('ERROR-HANDLING'))
        errorHandlingEnabled = None
        if errorHandling:
            if errorHandling == "TRANSFORMER-ERROR-HANDLING":
                errorHandlingEnabled = True
            elif errorHandling == "NO-TRANSFORMER-ERROR-HANDLING":
                errorHandlingEnabled = False
        return portRef, takeAddress, indirectAPI, errorHandlingEnabled

    def parseRunnableEntity(self, xmlRoot, parent):
        xmlDataReceivePoints = None
        xmlDataSendPoints = None
        xmlServerCallPoints = None
        xmlCanEnterExclusiveAreas = None
        adminData = None
        xmlModeAccessPoints = None
        xmlParameterAccessPoints = None
        canBeInvokedConcurrently = False
        xmlModeSwitchPoints = None
        minStartInterval = None
        if self.version < 4.0:
            for xmlElem in xmlRoot.findall('*'):
                if xmlElem.tag=='SHORT-NAME':
                    name=self.parseTextNode(xmlElem)
                elif xmlElem.tag=='CAN-BE-INVOKED-CONCURRENTLY':
                    canBeInvokedConcurrently=self.parseBooleanNode(xmlElem)
                elif xmlElem.tag=='DATA-RECEIVE-POINTS':
                    xmlDataReceivePoints=xmlElem
                elif xmlElem.tag=='DATA-SEND-POINTS':
                    xmlDataSendPoints=xmlElem
                elif xmlElem.tag=='SERVER-CALL-POINTS':
                    xmlServerCallPoints=xmlElem
                elif xmlElem.tag=='SYMBOL':
                    symbol=self.parseTextNode(xmlElem)
                elif xmlElem.tag=='CAN-ENTER-EXCLUSIVE-AREA-REFS':
                    xmlCanEnterExclusiveAreas=xmlElem
                elif xmlElem.tag=='ADMIN-DATA':
                    adminData=self.parseAdminDataNode(xmlElem)
                else:
                    raise NotImplementedError(xmlElem.tag)
        else:
            for xmlElem in xmlRoot.findall('*'):
                if xmlElem.tag=='SHORT-NAME':
                    name=self.parseTextNode(xmlElem)
                elif xmlElem.tag=='CAN-BE-INVOKED-CONCURRENTLY':
                    canBeInvokedConcurrently=self.parseBooleanNode(xmlElem)
                elif xmlElem.tag == 'MODE-ACCESS-POINTS':
                    xmlModeAccessPoints = xmlElem
                elif xmlElem.tag=='DATA-RECEIVE-POINT-BY-ARGUMENTS':
                    xmlDataReceivePoints=xmlElem
                elif xmlElem.tag=='DATA-SEND-POINTS':
                    xmlDataSendPoints=xmlElem
                elif xmlElem.tag=='SERVER-CALL-POINTS':
                    xmlServerCallPoints=xmlElem
                elif xmlElem.tag=='SYMBOL':
                    symbol=self.parseTextNode(xmlElem)
                elif xmlElem.tag=='CAN-ENTER-EXCLUSIVE-AREA-REFS':
                    xmlCanEnterExclusiveAreas=xmlElem
                elif xmlElem.tag == 'MINIMUM-START-INTERVAL':
                    minStartInterval = self.parseNumberNode(xmlElem)
                elif xmlElem.tag =='ADMIN-DATA':
                    adminData=self.parseAdminDataNode(xmlElem)
                elif xmlElem.tag == 'PARAMETER-ACCESSS':
                    xmlParameterAccessPoints = xmlElem
                elif xmlElem.tag == 'MODE-SWITCH-POINTS':
                    xmlModeSwitchPoints = xmlElem
                elif xmlElem.tag == 'READ-LOCAL-VARIABLES':
                    pass #implement later
                elif xmlElem.tag == 'WRITTEN-LOCAL-VARIABLES':
                    pass #implement later
                elif xmlElem.tag == 'DATA-READ-ACCESSS':
                    pass #implement later
                elif xmlElem.tag == 'DATA-WRITE-ACCESSS':
                    pass #implement later
                elif xmlElem.tag == 'RUNS-INSIDE-EXCLUSIVE-AREA-REFS':
                    pass #implement later
                else:
                    raise NotImplementedError(xmlElem.tag)
        runnableEntity = autosar.behavior.RunnableEntity(name, canBeInvokedConcurrently, symbol, parent)
        if minStartInterval is not None:
            runnableEntity.minStartInterval = float(1000 * minStartInterval)
        if xmlDataReceivePoints is not None:
            if self.version < 4.0:
                for xmlDataPoint in xmlDataReceivePoints.findall('./DATA-RECEIVE-POINT'):
                    name=self.parseTextNode(xmlDataPoint.find('SHORT-NAME'))
                    dataElementInstanceRef = self.parseDataElementInstanceRef(xmlDataPoint.find('DATA-ELEMENT-IREF'),'R-PORT-PROTOTYPE-REF')
                    if dataElementInstanceRef is not None:
                        dataReceivePoint=autosar.behavior.DataReceivePoint(dataElementInstanceRef.portRef,dataElementInstanceRef.dataElemRef,name)
                        runnableEntity.append(dataReceivePoint)
            else:
                for xmlVariableAcess in xmlDataReceivePoints.findall('VARIABLE-ACCESS'):
                    name=self.parseTextNode(xmlVariableAcess.find('SHORT-NAME'))
                    accessedVariable = self.parseAccessedVariable(xmlVariableAcess.find('./ACCESSED-VARIABLE'))
                    assert(accessedVariable is not None)
                    dataReceivePoint=autosar.behavior.DataReceivePoint(accessedVariable.portPrototypeRef,accessedVariable.targetDataPrototypeRef,name)
                    runnableEntity.append(dataReceivePoint)
        if xmlDataSendPoints is not None:
            if self.version < 4.0:
                for xmlDataPoint in xmlDataSendPoints.findall('./DATA-SEND-POINT'):
                    name=self.parseTextNode(xmlDataPoint.find('SHORT-NAME'))
                    dataElementInstanceRef = self.parseDataElementInstanceRef(xmlDataPoint.find('DATA-ELEMENT-IREF'),'P-PORT-PROTOTYPE-REF')
                    if dataElementInstanceRef is not None:
                        dataSendPoint=autosar.behavior.DataSendPoint(dataElementInstanceRef.portRef,dataElementInstanceRef.dataElemRef,name)
                        runnableEntity.append(dataSendPoint)
            else:
                for xmlVariableAcess in xmlDataSendPoints.findall('VARIABLE-ACCESS'):
                    name=self.parseTextNode(xmlVariableAcess.find('SHORT-NAME'))
                    accessedVariable = self.parseAccessedVariable(xmlVariableAcess.find('./ACCESSED-VARIABLE'))
                    assert(accessedVariable is not None)
                    dataSendPoint=autosar.behavior.DataSendPoint(accessedVariable.portPrototypeRef,accessedVariable.targetDataPrototypeRef,name)
                    runnableEntity.append(dataSendPoint)
        if xmlModeAccessPoints is not None:
            for xmlElem in xmlModeAccessPoints.findall('./*'):
                if xmlElem.tag == 'MODE-ACCESS-POINT':
                    modeAccessPoint = self.parseModeAccessPoint(xmlElem)
                    assert(modeAccessPoint is not None)
                    runnableEntity.modeAccessPoints.append(modeAccessPoint)
                else:
                    raise NotImplementedError(xmlElem.tag)
        if xmlServerCallPoints is not None:
            for xmlServerCallPoint in xmlServerCallPoints.findall('./SYNCHRONOUS-SERVER-CALL-POINT'):
                syncServerCallPoint = self.parseSyncServerCallPoint(xmlServerCallPoint)
                if syncServerCallPoint is not None: runnableEntity.serverCallPoints.append(syncServerCallPoint)
        if xmlCanEnterExclusiveAreas is not None:
            for xmlCanEnterExclusiveAreaRef in xmlCanEnterExclusiveAreas.findall('./CAN-ENTER-EXCLUSIVE-AREA-REF'):
                runnableEntity.exclusiveAreaRefs.append(self.parseTextNode(xmlCanEnterExclusiveAreaRef))
        if self.version >= 4.0 and xmlParameterAccessPoints is not None:
            for xmlChild in xmlParameterAccessPoints.findall('./*'):
                if xmlChild.tag == 'PARAMETER-ACCESS':
                    tmp = self.parseParameterAccessPoint(xmlChild, runnableEntity)
                    if tmp is not None:
                        runnableEntity.parameterAccessPoints.append(tmp)
                else:
                    raise NotImplementedError(xmlChild.tag)
        if xmlModeSwitchPoints is not None:
            for xmlElem in xmlModeSwitchPoints.findall('./*'):
                if xmlElem.tag == 'MODE-SWITCH-POINT':
                    modeSwitchPoint = self._parseModeSwitchPoint(xmlElem)
                    assert(modeSwitchPoint is not None)
                    runnableEntity.modeSwitchPoints.append(modeSwitchPoint)
                else:
                    raise NotImplementedError(xmlElem.tag)
        if runnableEntity is not None:
            runnableEntity.adminData = adminData
        return runnableEntity

    def parseModeAccessPoint(self, xmlRoot):
        assert(xmlRoot.tag == 'MODE-ACCESS-POINT')
        (name, modeGroupInstanceRef) = (None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'SHORT-NAME':
                name = self.parseTextNode(xmlElem)
            if xmlElem.tag == 'MODE-GROUP-IREF':
                for childElem in xmlElem.findall('./*'):
                    if childElem.tag == 'R-MODE-GROUP-IN-ATOMIC-SWC-INSTANCE-REF':
                        if modeGroupInstanceRef is None:
                            modeGroupInstanceRef=self._parseRequireModeGroupInstanceRef(childElem)
                        else:
                            raise NotImplementedError('Multiple instances of R-MODE-GROUP-IN-ATOMIC-SWC-INSTANCE-REF not implemented')
                    elif childElem.tag == 'P-MODE-GROUP-IN-ATOMIC-SWC-INSTANCE-REF':
                        if modeGroupInstanceRef is None:
                            modeGroupInstanceRef=self._parseProvideModeGroupInstanceRef(childElem)
                        else:
                            raise NotImplementedError('Multiple instances of P-MODE-GROUP-IN-ATOMIC-SWC-INSTANCE-REF not implemented')
                    else:
                        raise NotImplementedError(childElem.tag)
            else:
                raise NotImplementedError(xmlElem.tag)
        return autosar.behavior.ModeAccessPoint(name, modeGroupInstanceRef)

    def _parseModeSwitchPoint(self, xmlRoot):
        assert(xmlRoot.tag == 'MODE-SWITCH-POINT')
        (name, modeGroupInstanceRef) = (None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'SHORT-NAME':
                name = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'MODE-GROUP-IREF':
                modeGroupInstanceRef=self._parseProvideModeGroupInstanceRef(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        return autosar.behavior.ModeSwitchPoint(name, modeGroupInstanceRef)

    def parseParameterAccessPoint(self, xmlRoot, parent = None):
        assert(xmlRoot.tag == 'PARAMETER-ACCESS')
        (name, accessedParameter) = (None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'SHORT-NAME':
                name = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'ACCESSED-PARAMETER':
                for xmlChild in xmlElem.findall('./*'):
                    if xmlChild.tag == 'AUTOSAR-PARAMETER-IREF':
                        accessedParameter = self.parseParameterInstanceRef(xmlChild)
                    elif xmlChild.tag == 'LOCAL-PARAMETER-REF':
                        accessedParameter = autosar.behavior.LocalParameterRef(self.parseTextNode(xmlChild))
                    else:
                        raise NotImplementedError(xmlChild.tag)
            else:
                raise NotImplementedError(xmlElem.tag)
        if (name is not None) and (accessedParameter is not None):
            return autosar.behavior.ParameterAccessPoint(name, accessedParameter)

    def _parseRequireModeGroupInstanceRef(self, xmlRoot):
        """parses <R-MODE-GROUP-IN-ATOMIC-SWC-INSTANCE-REF>"""
        assert(xmlRoot.tag == 'R-MODE-GROUP-IN-ATOMIC-SWC-INSTANCE-REF')
        (requirePortRef, modeGroupRef ) = (None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'CONTEXT-R-PORT-REF':
                requirePortRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'TARGET-MODE-GROUP-REF':
                modeGroupRef = self.parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        if requirePortRef is None:
            raise RuntimeError('CONTEXT-R-PORT-REF not set')
        if modeGroupRef is None:
            raise RuntimeError('TARGET-MODE-GROUP-REF not set')
        return autosar.behavior.RequireModeGroupInstanceRef(requirePortRef, modeGroupRef)

    def _parseProvideModeGroupInstanceRef(self, xmlRoot):
        """parses <P-MODE-GROUP-IN-ATOMIC-SWC-INSTANCE-REF>"""
        #This XML item exists multiple times (at least 4 different places) in the AUTOSAR 4 XSD using different XML tag.
        assert(xmlRoot.tag in ['P-MODE-GROUP-IN-ATOMIC-SWC-INSTANCE-REF', 'MODE-GROUP-IREF'])
        (providePortRef, modeGroupRef ) = (None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'CONTEXT-P-PORT-REF':
                providePortRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'TARGET-MODE-GROUP-REF':
                modeGroupRef = self.parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        if providePortRef is None:
            raise RuntimeError('CONTEXT-P-PORT-REF not set')
        if modeGroupRef is None:
            raise RuntimeError('TARGET-MODE-GROUP-REF not set')
        return autosar.behavior.ProvideModeGroupInstanceRef(providePortRef, modeGroupRef)


    def parseModeInstanceRef(self, xmlRoot):
        """parses <MODE-IREF>"""
        assert(xmlRoot.tag == 'MODE-IREF')
        if self.version < 4.0:
            modeDeclarationRef=self.parseTextNode(xmlRoot.find('MODE-DECLARATION-REF'))
            modeDeclarationGroupPrototypeRef = self.parseTextNode(xmlRoot.find('MODE-DECLARATION-GROUP-PROTOTYPE-REF'))
            requirePortPrototypeRef = self.parseTextNode(xmlRoot.find('R-PORT-PROTOTYPE-REF'))
        elif self.version >= 4.0:
            modeDeclarationRef=self.parseTextNode(xmlRoot.find('TARGET-MODE-DECLARATION-REF'))
            modeDeclarationGroupPrototypeRef = self.parseTextNode(xmlRoot.find('CONTEXT-MODE-DECLARATION-GROUP-PROTOTYPE-REF'))
            requirePortPrototypeRef = self.parseTextNode(xmlRoot.find('CONTEXT-PORT-REF'))
        else:
            raise NotImplemented('version: '+self.version)
        return autosar.behavior.ModeInstanceRef(modeDeclarationRef,modeDeclarationGroupPrototypeRef,requirePortPrototypeRef)

    def parseParameterInstanceRef(self, xmlRoot):
        """parses <AUTOSAR-PARAMETER-IREF>"""
        assert(xmlRoot.tag == 'AUTOSAR-PARAMETER-IREF')
        portRef = self.parseTextNode(xmlRoot.find('PORT-PROTOTYPE-REF'))
        parameterDataRef = self.parseTextNode(xmlRoot.find('TARGET-DATA-PROTOTYPE-REF'))
        return autosar.behavior.ParameterInstanceRef(portRef, parameterDataRef)

    def _parseModeDependency(self,xmlRoot,parent=None):
        """parses <MODE-DEPENDENCY>"""
        assert(xmlRoot.tag == 'MODE-DEPENDENCY')
        modeDependency=autosar.behavior.ModeDependency()
        if xmlRoot.find('DEPENDENT-ON-MODE-IREFS') is not None:
            for xmlNode in xmlRoot.findall('./DEPENDENT-ON-MODE-IREFS/DEPENDENT-ON-MODE-IREF'):
                modeInstanceRef = self._parseDependentOnModeInstanceRef(xmlNode)
                if modeInstanceRef is not None:
                    modeDependency.modeInstanceRefs.append(modeInstanceRef)
        return modeDependency

    def _parseDependentOnModeInstanceRef(self,xmlRoot,parent=None):
        """parses <DEPENDENT-ON-MODE-IREF>"""
        assert(xmlRoot.tag == 'DEPENDENT-ON-MODE-IREF')
        modeDeclarationRef=self.parseTextNode(xmlRoot.find('MODE-DECLARATION-REF'))
        modeDeclarationGroupPrototypeRef = self.parseTextNode(xmlRoot.find('MODE-DECLARATION-GROUP-PROTOTYPE-REF'))
        requirePortPrototypeRef = self.parseTextNode(xmlRoot.find('R-PORT-PROTOTYPE-REF'))
        return autosar.behavior.ModeDependencyRef(modeDeclarationRef,modeDeclarationGroupPrototypeRef,requirePortPrototypeRef)

    def _parseDisabledModesInstanceRefs(self,xmlRoot,parent=None):
        """parses <DISABLED-MODE-IREFS>"""
        assert(xmlRoot.tag == 'DISABLED-MODE-IREFS')
        disabledInModes = []
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'DISABLED-MODE-IREF':
                disabledInModes.append(self._parseDisabledModeInstanceRef(xmlElem))
            else:
                raise NotImplementedError(xmlElem.tag)
        return disabledInModes

    def _parseDisabledModeInstanceRef(self, xmlRoot):
        """parses <DISABLED-MODE-IREF>"""
        (requirePortPrototypeRef, modeDeclarationGroupPrototypeRef, modeDeclarationRef) = (None, None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'CONTEXT-PORT-REF':
                requirePortPrototypeRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'CONTEXT-MODE-DECLARATION-GROUP-PROTOTYPE-REF':
                modeDeclarationGroupPrototypeRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'TARGET-MODE-DECLARATION-REF':
                modeDeclarationRef = self.parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        if (modeDeclarationRef is not None) and (modeDeclarationGroupPrototypeRef is not None) and (requirePortPrototypeRef is not None):
            return autosar.behavior.DisabledModeInstanceRef(modeDeclarationRef,modeDeclarationGroupPrototypeRef,requirePortPrototypeRef)
        else:
            raise RuntimeError('Parse Error: <CONTEXT-PORT-REF DEST>, <CONTEXT-MODE-DECLARATION-GROUP-PROTOTYPE-REF> and <TARGET-MODE-DECLARATION-REF> must be defined')

    def parseInitEvent(self,xmlNode,parent=None):
        name = self.parseTextNode(xmlNode.find('SHORT-NAME'))
        startOnEventRef = self.parseTextNode(xmlNode.find('START-ON-EVENT-REF'))
        initEvent=autosar.behavior.InitEvent(name, startOnEventRef, parent)
        return initEvent

    def parseModeSwitchEvent(self,xmlNode,parent=None):
        """parses AUTOSAR3 <MODE-SWITCH-EVENT>"""
        if self.version < 4.0:
            assert(xmlNode.tag=='MODE-SWITCH-EVENT')
            name = self.parseTextNode(xmlNode.find('SHORT-NAME'))
            modeInstRef = self.parseModeInstanceRef(xmlNode.find('MODE-IREF'))
            startOnEventRef = self.parseTextNode(xmlNode.find('START-ON-EVENT-REF'))
            activation = self.parseTextNode(xmlNode.find('ACTIVATION'))
            modeSwitchEvent=autosar.behavior.ModeSwitchEvent(name, startOnEventRef, activation, parent, self.version)
            modeSwitchEvent.modeInstRef=modeInstRef
        elif self.version >= 4.0:
            assert(xmlNode.tag=='SWC-MODE-SWITCH-EVENT')
            name = self.parseTextNode(xmlNode.find('SHORT-NAME'))
            modeInstanceRefs = []
            for xmlElem in xmlNode.findall('./MODE-IREFS/*'):
                if xmlElem.tag == 'MODE-IREF':
                    modeInstanceRefs.append(self.parseModeInstanceRef(xmlElem))
                else:
                    raise NotImplementedError(xmlElem.tag)
            startOnEventRef = self.parseTextNode(xmlNode.find('START-ON-EVENT-REF'))
            activation = self.parseTextNode(xmlNode.find('ACTIVATION'))
            modeSwitchEvent=autosar.behavior.ModeSwitchEvent(name, startOnEventRef, activation, parent, self.version)
            modeSwitchEvent.modeInstRef=modeInstanceRefs[0]
        else:
            raise NotImplementedError('version: '+self.version)
        return modeSwitchEvent

    def parseModeSwitchedAckEvent(self, xmlNode, parent = None):
        if self.version < 4.0:
            raise NotImplementedError(xmlNode.tag)
        else:
            assert(xmlNode.tag=='MODE-SWITCHED-ACK-EVENT')
            (name, startOnEventRef, eventSourceRef, xmlDisabledModeRefs) = (None, None, None, None)
            for xmlElem in xmlNode.findall('./*'):
                if xmlElem.tag == 'SHORT-NAME':
                    name = self.parseTextNode(xmlElem)
                elif xmlElem.tag == 'DISABLED-MODE-IREFS':
                    xmlDisabledModeRefs = xmlElem
                elif xmlElem.tag == 'START-ON-EVENT-REF':
                    startOnEventRef = self.parseTextNode(xmlElem)
                elif xmlElem.tag == 'EVENT-SOURCE-REF':
                    eventSourceRef = self.parseTextNode(xmlElem)
                else:
                    raise NotImplementedError(xmlElem.tag)
            modeSwitchEvent = autosar.behavior.ModeSwitchAckEvent(name, startOnEventRef, eventSourceRef, parent)
            if xmlDisabledModeRefs is not None:
                modeSwitchEvent.disabledInModes = self._parseDisabledModesInstanceRefs(xmlDisabledModeRefs, parent)
            return modeSwitchEvent

    def parseTimingEvent(self,xmlNode,parent=None):
        (name, startOnEventRef, period, xmlModeDepency, xmlDisabledModeRefs) = (None, None, None, None, None)
        for xmlElem in xmlNode.findall('./*'):
            if xmlElem.tag == 'SHORT-NAME':
                name = self.parseTextNode(xmlElem)
            elif self.version >= 4.0 and xmlElem.tag == 'DISABLED-MODE-IREFS':
                xmlDisabledModeRefs =xmlElem
            elif self.version < 4.0 and xmlElem.tag == 'MODE-DEPENDENCY':
                xmlModeDepency =xmlElem
            elif xmlElem.tag == 'START-ON-EVENT-REF':
                startOnEventRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'PERIOD':
                period = self.parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)

        if (name is not None) and (startOnEventRef is not None) and (period is not None):
            timingEvent=autosar.behavior.TimingEvent(name, startOnEventRef, float(period)*1000, parent)
            if xmlModeDepency is not None:
                timingEvent.modeDependency = self._parseModeDependency(xmlModeDepency, parent)
            elif xmlDisabledModeRefs is not None:
                timingEvent.disabledInModes = self._parseDisabledModesInstanceRefs(xmlDisabledModeRefs, parent)
            return timingEvent
        else:
            raise RuntimeError('Parse error: <SHORT-NAME> and <START-ON-EVENT-REF> and <PERIOD> must be defined')

    def parseDataReceivedEvent(self,xmlRoot,parent=None):
        name = self.parseTextNode(xmlRoot.find('SHORT-NAME'))
        startOnEventRef = self.parseTextNode(xmlRoot.find('START-ON-EVENT-REF'))
        portTag = 'CONTEXT-R-PORT-REF' if self.version >= 4.0 else 'R-PORT-PROTOTYPE-REF'
        dataInstanceRef=self.parseDataInstanceRef(xmlRoot.find('DATA-IREF'),portTag)
        dataReceivedEvent=autosar.behavior.DataReceivedEvent(name, startOnEventRef, parent)
        xmlModeDependency = xmlRoot.find('MODE-DEPENDENCY')
        if xmlModeDependency is not None:
            dataReceivedEvent.modeDependency = self._parseModeDependency(xmlModeDependency)
        dataReceivedEvent.dataInstanceRef=dataInstanceRef
        return dataReceivedEvent

    def parseOperationInvokedEvent(self,xmlRoot,parent=None):
        (name, startOnEventRef, modeDependency, operationInstanceRef) = (None, None, None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'SHORT-NAME':
                name = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'START-ON-EVENT-REF':
                startOnEventRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'OPERATION-IREF':
                portTag = 'CONTEXT-P-PORT-REF' if self.version >= 4.0 else 'P-PORT-PROTOTYPE-REF'
                operationInstanceRef = self.parseOperationInstanceRef(xmlElem, portTag)
            elif xmlElem.tag == 'MODE-DEPENDENCY':
                modeDependency = self._parseModeDependency(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        operationInvokedEvent=autosar.behavior.OperationInvokedEvent(name, startOnEventRef, parent)
        operationInvokedEvent.modeDependency = modeDependency
        operationInvokedEvent.operationInstanceRef = operationInstanceRef
        return operationInvokedEvent

    def parseDataInstanceRef(self, xmlRoot, portTag):
        """parses <DATA-IREF>"""
        assert(xmlRoot.tag=='DATA-IREF')
        dataElemTag = 'TARGET-DATA-ELEMENT-REF' if self.version >= 4.0 else 'DATA-ELEMENT-PROTOTYPE-REF'
        (portRef, dataElemRef) = (None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == portTag:
                portRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == dataElemTag:
                dataElemRef = self.parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        return autosar.behavior.DataInstanceRef(portRef,dataElemRef)

    def parseOperationInstanceRef(self,xmlRoot,portTag):
        """parses <OPERATION-IREF>"""
        assert(xmlRoot.tag=='OPERATION-IREF')
        assert(xmlRoot.find(portTag) is not None)

        if self.version >= 4.0:
            if portTag == 'CONTEXT-P-PORT-REF':
                operationTag = 'TARGET-PROVIDED-OPERATION-REF'
            else:
                operationTag = 'TARGET-REQUIRED-OPERATION-REF'
        else:
            operationTag = 'OPERATION-PROTOTYPE-REF'
        return autosar.behavior.OperationInstanceRef(self.parseTextNode(xmlRoot.find(portTag)),self.parseTextNode(xmlRoot.find(operationTag)))


    def parseDataElementInstanceRef(self,xmlRoot,portTag):
        """parses <DATA-ELEMENT-IREF>"""
        assert(xmlRoot.tag=='DATA-ELEMENT-IREF')
        assert(xmlRoot.find(portTag) is not None)
        return autosar.behavior.DataElementInstanceRef(self.parseTextNode(xmlRoot.find(portTag)),self.parseTextNode(xmlRoot.find('DATA-ELEMENT-PROTOTYPE-REF')))

    def parseSwcNvBlockNeeds(self,xmlRoot):
        name=self.parseTextNode(xmlRoot.find('SHORT-NAME'))
        numberOfDataSets=self.parseIntNode(xmlRoot.find('N-DATA-SETS'))
        readOnly=self.parseBooleanNode(xmlRoot.find('READONLY'))
        reliability=self.parseTextNode(xmlRoot.find('RELIABILITY'))
        resistantToChangedSW=self.parseBooleanNode(xmlRoot.find('RESISTANT-TO-CHANGED-SW'))
        restoreAtStart=self.parseBooleanNode(xmlRoot.find('RESTORE-AT-START'))
        writeOnlyOnce=self.parseBooleanNode(xmlRoot.find('WRITE-ONLY-ONCE'))
        writingFrequency=self.parseIntNode(xmlRoot.find('WRITING-FREQUENCY'))
        writingPriority=self.parseTextNode(xmlRoot.find('WRITING-PRIORITY'))
        defaultBlockRef=self.parseTextNode(xmlRoot.find('DEFAULT-BLOCK-REF'))
        mirrorBlockRef=self.parseTextNode(xmlRoot.find('MIRROR-BLOCK-REF'))
        serviceCallPorts=self.parseServiceCallPorts(xmlRoot.find('SERVICE-CALL-PORTS'))
        assert(len(serviceCallPorts)>0)
        swcNvBlockNeeds=autosar.behavior.SwcNvBlockNeeds(name,numberOfDataSets,readOnly,reliability,resistantToChangedSW,restoreAtStart,
                                        writeOnlyOnce,writingFrequency,writingPriority,defaultBlockRef,mirrorBlockRef)
        swcNvBlockNeeds.serviceCallPorts=serviceCallPorts
        return swcNvBlockNeeds

    def parseServiceCallPorts(self,xmlRoot):
        """parses <SERVICE-CALL-PORTS>"""
        assert(xmlRoot.tag=='SERVICE-CALL-PORTS')
        serviceCallPorts=[]
        for xmlNode in xmlRoot.findall('ROLE-BASED-R-PORT-ASSIGNMENT'):
            roleBasedRPortAssignment=autosar.behavior.RoleBasedRPortAssignment(self.parseTextNode(xmlNode.find('R-PORT-PROTOTYPE-REF')),self.parseTextNode(xmlNode.find('ROLE')))
            serviceCallPorts.append(roleBasedRPortAssignment)
        return serviceCallPorts

    def parseCalPrmElemPrototype(self, xmlRoot, parent):
        """
        parses <CALPRM-ELEMENT-PROTOTYPE>
        """
        name = self.parseTextNode(xmlRoot.find('SHORT-NAME'))
        adminData=self.parseAdminDataNode(xmlRoot.find('ADMIN-DATA'))
        typeRef = self.parseTextNode(xmlRoot.find('TYPE-TREF'))
        calPrmElemPrototype = autosar.behavior.CalPrmElemPrototype(name, typeRef, parent, adminData)
        for xmlElem in xmlRoot.findall('./SW-DATA-DEF-PROPS/*'):
            if xmlElem.tag=='SW-ADDR-METHOD-REF':
                calPrmElemPrototype.swDataDefsProps.append(self.parseTextNode(xmlElem))
            else:
                raise NotImplementedError(xmlElem.tag)
        return calPrmElemPrototype

    def parseSyncServerCallPoint(self, xmlRoot):
        """
        parses <SYNCHRONOUS-SERVER-CALL-POINT>
        """
        assert(xmlRoot.tag=='SYNCHRONOUS-SERVER-CALL-POINT')
        timeout=0.0
        if self.version >= 4.0:
            operationInstanceRefs=[]
            for xmlElem in xmlRoot.findall('*'):
                if xmlElem.tag=='SHORT-NAME':
                    name=self.parseTextNode(xmlElem)
                elif xmlElem.tag=='OPERATION-IREF':
                    operationInstanceRefs.append(self.parseOperationInstanceRef(xmlElem,'CONTEXT-R-PORT-REF'))
                elif xmlElem.tag=='TIMEOUT':
                    timeout=self.parseFloatNode(xmlElem)
                else:
                    raise NotImplementedError(xmlElem.tag)
        else:
            for xmlElem in xmlRoot.findall('*'):
                if xmlElem.tag=='SHORT-NAME':
                    name=self.parseTextNode(xmlElem)
                elif xmlElem.tag=='OPERATION-IREFS':
                    operationInstanceRefs=[]
                    for xmlOperation in xmlElem.findall('*'):
                        if xmlOperation.tag=='OPERATION-IREF':
                            operationInstanceRefs.append(self.parseOperationInstanceRef(xmlOperation,'R-PORT-PROTOTYPE-REF'))
                        else:
                            raise NotImplementedError(xmlElem.tag)
                elif xmlElem.tag=='TIMEOUT':
                    timeout=self.parseFloatNode(xmlElem)
                else:
                    raise NotImplementedError(xmlElem.tag)
        retval=autosar.behavior.SyncServerCallPoint(name,timeout)
        retval.operationInstanceRefs=operationInstanceRefs
        return retval

    def parseAccessedVariable(self, xmlRoot):
        assert(xmlRoot.tag == 'ACCESSED-VARIABLE')
        xmlPortPrototypeRef = xmlRoot.find('./AUTOSAR-VARIABLE-IREF/PORT-PROTOTYPE-REF')
        xmlTargetDataPrototypeRef = xmlRoot.find('./AUTOSAR-VARIABLE-IREF/TARGET-DATA-PROTOTYPE-REF')
        assert (xmlPortPrototypeRef is not None)
        assert (xmlTargetDataPrototypeRef is not None)
        return autosar.behavior.VariableAccess(self.parseTextNode(xmlRoot.find('SHORT-NAME')),self.parseTextNode(xmlPortPrototypeRef), self.parseTextNode(xmlTargetDataPrototypeRef))

    def parseSwcServiceDependency(self, xmlRoot, parent = None):
        """parses <SWC-SERVICE-DEPENDENCY>"""
        assert(xmlRoot.tag == 'SWC-SERVICE-DEPENDENCY')
        (name, desc, serviceNeeds) = (None, None, None)
        roleBasedDataAssignments = []
        roleBasedPortAssignments = []
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'SHORT-NAME':
                name = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'DESC':
                desc = self.parseDescDirect(xmlElem)
            elif xmlElem.tag == 'ASSIGNED-DATAS':
                for xmlChildElem in xmlElem.findall('./*'):
                    if xmlChildElem.tag == 'ROLE-BASED-DATA-ASSIGNMENT':
                        tmp = self._parseRoleBasedDataAssignment(xmlChildElem)
                        if tmp is not None: roleBasedDataAssignments.append(tmp)
                    else:
                        raise NotImplementedError(xmlChildElem.tag)
            elif xmlElem.tag == 'ASSIGNED-PORTS':
                for xmlChildElem in xmlElem.findall('./*'):
                    if xmlChildElem.tag == 'ROLE-BASED-PORT-ASSIGNMENT':
                        tmp = self._parseRoleBasedPortAssignment(xmlChildElem)
                        if tmp is not None: roleBasedPortAssignments.append(tmp)
                    else:
                        raise NotImplementedError(xmlChildElem.tag)
            elif xmlElem.tag == 'SERVICE-NEEDS':
                serviceNeeds = self.parseServiceNeeds(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        swcServiceDependency = autosar.behavior.SwcServiceDependency(name, parent = parent)
        if desc is not None:
            swcServiceDependency.desc=desc[0]
            swcServiceDependency.descAttr=desc[1]
        if serviceNeeds is not None:
            swcServiceDependency.serviceNeeds = serviceNeeds
        if len(roleBasedDataAssignments) > 0:
            swcServiceDependency.roleBasedDataAssignments = roleBasedDataAssignments
        if len(roleBasedPortAssignments) > 0:
            swcServiceDependency.roleBasedPortAssignments = roleBasedPortAssignments
        return swcServiceDependency


    def parseParameterDataPrototype(self, xmlRoot, parent = None):
        """parses <PARAMETER-DATA-PROTOTYPE> (AUTOSAR 4)"""
        assert(xmlRoot.tag == 'PARAMETER-DATA-PROTOTYPE')
        (variants, typeRef, swAddressMethodRef, swCalibrationAccess, initValue, initValueRef) = (None, None, None, None, None, None)
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'SW-DATA-DEF-PROPS':
                variants = self.parseSwDataDefProps(xmlElem)
                if len(variants) > 0:
                    swAddressMethodRef = variants[0].swAddressMethodRef
                    swCalibrationAccess = variants[0].swCalibrationAccess
            elif xmlElem.tag == 'INIT-VALUE':
                for xmlChild in xmlElem.findall('./*'):
                    if xmlChild.tag == 'CONSTANT-REFERENCE':
                        initValueRef = self.parseTextNode(xmlChild.find('./CONSTANT-REF'))
                    else:
                        values = self.constantParser.parseValueV4(xmlElem, None)
                        if len(values) != 1:
                            raise ValueError('{0} cannot cannot contain multiple elements'.format(xmlElem.tag))
                        initValue = values[0]
            elif xmlElem.tag == 'TYPE-TREF':
                typeRef = self.parseTextNode(xmlElem)
            else:
                self.baseHandler(xmlElem)
        obj = autosar.element.ParameterDataPrototype(self.name, typeRef=typeRef,
                    swAddressMethodRef=swAddressMethodRef, swCalibrationAccess=swCalibrationAccess,
                    initValue=initValue, initValueRef=initValueRef, parent=parent, adminData=self.adminData)
        self.pop(obj)
        return obj

    def parseServiceNeeds(self, xmlRoot, parent = None):
        """parses <SERVICE-NEEDS> (AUTOSAR 4)"""
        assert(xmlRoot.tag == 'SERVICE-NEEDS')
        (xmlNvBlockNeeds, nvBlockNeeds) = (None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'NV-BLOCK-NEEDS':
                xmlNvBlockNeeds = xmlElem
            else:
                raise NotImplementedError(xmlElem.tag)
        serviceNeeds = autosar.behavior.ServiceNeeds(None, parent)
        if xmlNvBlockNeeds is not None:
            serviceNeeds.nvmBlockNeeds = self.parseNvmBlockNeeds(xmlElem, serviceNeeds)
        return serviceNeeds

    def parseNvmBlockNeeds(self, xmlRoot, parent = None):
        """parses <NV-BLOCK-NEEDS> (AUTOSAR 4)"""
        config = autosar.behavior.NvmBlockConfig()

        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'N-DATA-SETS':
                config.numberOfDataSets = self.parseIntNode(xmlElem)
            elif xmlElem.tag == 'N-ROM-BLOCKS':
                config.numberOfRomBlocks = self.parseIntNode(xmlElem)
            elif xmlElem.tag == 'RAM-BLOCK-STATUS-CONTROL':
                config.ramBlockStatusControl = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'RELIABILITY':
                config.reliability = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'WRITING-PRIORITY':
                config.writingPriority = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'WRITING-FREQUENCY':
                config.writingFrequency = self.parseIntNode(xmlElem)
            elif xmlElem.tag == 'CALC-RAM-BLOCK-CRC':
                config.calcRamBlockCrc = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'CHECK-STATIC-BLOCK-ID':
                config.checkStaticBlockId = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'READONLY':
                config.readOnly = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'RESISTANT-TO-CHANGED-SW':
                config.resistantToChangedSw = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'RESTORE-AT-START':
                config.restoreAtStartup = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'STORE-AT-SHUTDOWN':
                config.storeAtShutdown = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'WRITE-VERIFICATION':
                config.writeVerification = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'WRITE-ONLY-ONCE':
                config.writeOnlyOnce = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'USE-AUTO-VALIDATION-AT-SHUT-DOWN':
                config.autoValidationAtShutdown = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'USE-CRC-COMP-MECHANISM':
                config.useCrcCompMechanism = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'STORE-EMERGENCY':
                config.storeEmergency = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'STORE-IMMEDIATE':
                config.storeImmediate = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'STORE-CYCLIC':
                config.storeCyclic = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'CYCLIC-WRITING-PERIOD':
                config.cyclicWritePeriod = self.parseFloatNode(xmlElem)
            else:
                self.baseHandler(xmlElem)

        if self.name is None:
            raise RuntimeError('<SHORT-NAME> is missing or incorrectly formatted')
        config.check()
        obj = autosar.behavior.NvmBlockNeeds(self.name, blockConfig = config, parent = parent, adminData = self.adminData)
        self.pop(obj)
        return obj

    def _parseRoleBasedDataAssignment(self, xmlRoot):
        assert(xmlRoot.tag == 'ROLE-BASED-DATA-ASSIGNMENT')
        (role, localVariableRef, localParameterRef) = (None, None, None)

        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'ROLE':
                role = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'USED-DATA-ELEMENT':
                for xmlChild in xmlElem.findall('./*'):
                    if xmlChild.tag == 'LOCAL-VARIABLE-REF':
                        localVariableRef = self.parseTextNode(xmlChild)
                    else:
                        raise NotImplementedError(xmlChild.tag)
            elif xmlElem.tag == 'USED-PARAMETER-ELEMENT':
                pass
                for xmlChild in xmlElem.findall('./*'):
                    if xmlChild.tag == 'LOCAL-PARAMETER-REF':
                        localParameterRef = autosar.behavior.LocalParameterRef(self.parseTextNode(xmlChild))
                    else:
                        raise NotImplementedError(xmlChild.tag)
            else:
                raise NotImplementedError(xmlElem.tag)
        if (role is not None) and ( (localVariableRef is not None) or (localParameterRef is not None) ):
            return autosar.behavior.RoleBasedDataAssignment(role, localVariableRef, localParameterRef)
        return None


    def _parseRoleBasedPortAssignment(self, xmlRoot):
        assert(xmlRoot.tag == 'ROLE-BASED-PORT-ASSIGNMENT')
        (portRef, role) = (None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'PORT-PROTOTYPE-REF':
                portRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'ROLE':
                portRef = self.parseTextNode(xmlElem)
            else:
                self.baseHandler(xmlElem)
        if portRef is not None:
            return autosar.behavior.RoleBasedPortAssignment(portRef, role)
        return None

    def parseAutosarVariableRefXML(self, xmlRoot):
        (localVariableRef, portPrototypeRef, targetDataPrototypeRef) = (None, None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'LOCAL-VARIABLE-REF':
                localVariableRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'AUTOSAR-VARIABLE-IREF' or xmlElem.tag == 'AUTOSAR-VARIABLE-IN-IMPL-DATATYPE':
                xmlPortPrototypeRef = xmlElem.find('./PORT-PROTOTYPE-REF')
                xmlTargetDataPrototypeRef = xmlElem.find('./TARGET-DATA-PROTOTYPE-REF')
                assert (xmlPortPrototypeRef is not None)
                assert (xmlTargetDataPrototypeRef is not None)
                portPrototypeRef = self.parseTextNode(xmlPortPrototypeRef)
                targetDataPrototypeRef = self.parseTextNode(xmlTargetDataPrototypeRef)
        return localVariableRef, portPrototypeRef, targetDataPrototypeRef

    def parseNvBlockSWCnvBlockDescriptor(self, xmlRoot, parent):
        """AUTOSAR 4 NV-BLOCK-DESCRIPTOR"""
        assert(xmlRoot.tag == 'NV-BLOCK-DESCRIPTOR')
        name = self.parseTextNode(xmlRoot.find('SHORT-NAME'))
        ws = parent.rootWS()
        assert(ws is not None)
        if (name is not None):
            handledXML = ['SHORT-NAME']
            descriptor = autosar.behavior.NvBlockDescriptor(name, parent)
            for xmlElem in xmlRoot.findall('./*'):
                if xmlElem.tag in handledXML:
                    pass
                elif xmlElem.tag == 'DATA-TYPE-MAPPING-REFS':
                    for xmlChild in xmlElem.findall('./*'):
                        if xmlChild.tag == 'DATA-TYPE-MAPPING-REF':
                            tmp = self.parseTextNode(xmlChild)
                            assert(tmp is not None)
                            descriptor.dataTypeMappingRefs.append(tmp)
                elif xmlElem.tag == 'NV-BLOCK-DATA-MAPPINGS':
                    for xmlMapping in xmlElem.findall('./NV-BLOCK-DATA-MAPPING'):
                        dataMapping = autosar.behavior.NvBlockDataMapping(descriptor)
                        descriptor.nvBlockDataMappings.append(dataMapping)
                        for xmlData in xmlMapping.findall('./*'):
                            localVariableRef, portPrototypeRef, targetDataPrototypeRef = self.parseAutosarVariableRefXML(xmlData)
                            if xmlData.tag == 'NV-RAM-BLOCK-ELEMENT':
                                if localVariableRef is None:
                                    raise ValueError('Cannot find needed LOCAL-VARIABLE-REF for NV-RAM-BLOCK-ELEMENT in {0}'.format(descriptor.name))
                                dataMapping.nvRamBlockElement = autosar.behavior.NvRamBlockElement(dataMapping, localVariableRef=localVariableRef)
                            elif xmlData.tag == 'READ-NV-DATA':
                                if portPrototypeRef is None and targetDataPrototypeRef is None:
                                    raise ValueError('Cannot find needed AUTOSAR-VARIABLE-IREF or AUTOSAR-VARIABLE-IN-IMPL-DATATYPE for READ-NV-DATA in {0}'.format(descriptor.name))
                                dataMapping.readNvData = autosar.behavior.ReadNvData(dataMapping, autosarVariablePortRef=portPrototypeRef, autosarVariableElementRef=targetDataPrototypeRef)
                            elif xmlData.tag == 'WRITTEN-NV-DATA':
                                if portPrototypeRef is None and targetDataPrototypeRef is None:
                                    raise ValueError('Cannot find needed AUTOSAR-VARIABLE-IREF or AUTOSAR-VARIABLE-IN-IMPL-DATATYPE for WRITTEN-NV-DATA in {0}'.format(descriptor.name))
                                dataMapping.writtenNvData = autosar.behavior.WrittenNvData(dataMapping, autosarVariablePortRef=portPrototypeRef, autosarVariableElementRef=targetDataPrototypeRef)
                            elif xmlData.tag == 'WRITTEN-READ-NV-DATA':
                                if portPrototypeRef is None and targetDataPrototypeRef is None:
                                    raise ValueError('Cannot find needed AUTOSAR-VARIABLE-IREF or AUTOSAR-VARIABLE-IN-IMPL-DATATYPE for WRITTEN-READ-NV-DATA in {0}'.format(descriptor.name))
                                dataMapping.writtenReadNvData = autosar.behavior.WrittenReadNvData(dataMapping, autosarVariablePortRef=portPrototypeRef, autosarVariableElementRef=targetDataPrototypeRef)
                            else:
                                raise NotImplementedError(xmlData.tag)
                elif xmlElem.tag == 'NV-BLOCK-NEEDS':
                    descriptor.nvBlockNeeds = self.parseNvmBlockNeeds(xmlElem, descriptor)
                elif xmlElem.tag == 'RAM-BLOCK':
                    # Change tag so it is correct for the parser.
                    xmlElem.tag = 'VARIABLE-DATA-PROTOTYPE'
                    dataElement = self.parseVariableDataPrototype(xmlElem, descriptor)
                    # Cast the object to correct class.
                    descriptor.ramBlock = autosar.behavior.NvBlockRamBlock.cast(dataElement)
                elif xmlElem.tag == 'ROM-BLOCK':
                    # Change tag so it is correct for the parser.
                    xmlElem.tag = 'PARAMETER-DATA-PROTOTYPE'
                    dataElement = self.parseParameterDataPrototype(xmlElem, descriptor)
                    # Cast the object to correct class.
                    descriptor.romBlock = autosar.behavior.NvBlockRomBlock.cast(dataElement)
                elif xmlElem.tag == 'SUPPORT-DIRTY-FLAG':
                    dirtyFlag = self.parseBooleanNode(xmlElem)
                    if dirtyFlag is not None:
                        descriptor.supportDirtyFlag = dirtyFlag
                elif xmlElem.tag == 'TIMING-EVENT-REF':
                    timingRef = self.parseTextNode(xmlElem)
                    parts = timingRef.partition('/')
                    while True:
                        if parts[0] == parent.name:
                            timingEvent = parent.find(parts[2])
                            if timingEvent is not None:
                                break
                        elif len(parts[2]) == 0:
                            break
                        parts = parts[2].partition('/')

                    if timingEvent is None:
                        raise ValueError('Cannot find timing event {0}'.format(timingRef))
                    descriptor.timingEventRef = timingEvent.name
                else:
                    raise NotImplementedError(xmlElem.tag)
            return descriptor



from autosar.writer.writer_base import BaseWriter
from autosar.writer.package_writer import PackageWriter
from autosar.base import applyFilter
import collections

class WorkspaceWriter(BaseWriter):
    def __init__(self, version, patch, schema, packageWriter):
        super().__init__(version, patch)
        assert(isinstance(packageWriter, PackageWriter))
        self.schema = schema
        self.packageWriter=packageWriter

    def beginFile(self):
        lines=[]
        if (self.version >= 3.0) and (self.version < 4.0):
            lines.append('<?xml version="1.0" encoding="UTF-8"?>')
            versionString = "%d.%d.%d"%(self.major, self.minor, self.patch)
            if self.schema is None:
                schema = 'autosar_'+versionString+'.xsd'
            else:
                schema = self.schema
            lines.append('<AUTOSAR xsi:schemaLocation="http://autosar.org/%s %s" xmlns="http://autosar.org/%s" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'%(versionString, schema, versionString))
            lines.append(self.indentChar+'<TOP-LEVEL-PACKAGES>')
        elif self.version >= 4.0:
            lines.append('<?xml version="1.0" encoding="utf-8"?>')
            lines.append('<AUTOSAR xmlns="http://autosar.org/schema/r4.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://autosar.org/schema/r4.0 AUTOSAR_{0}-{1}-{2}.xsd">'.format(self.major, self.minor, self.patch))
            lines.append(self.indentChar+'<AR-PACKAGES>')
        return lines

    def endFile(self):
        lines=[]
        if (self.version >= 3.0) and (self.version<4.0):
            lines.append(self.indentChar+'</TOP-LEVEL-PACKAGES>')
        else:
            lines.append(self.indentChar+'</AR-PACKAGES>')
        lines.append('</AUTOSAR>')
        return lines


    def saveXML(self, ws, fp, filters, ignore):
        fp.write(self.toXML(ws, filters, ignore))

    def toXML(self, ws, filters, ignore):
        lines=self.beginFile()
        result='\n'.join(lines)+'\n'
        for package in ws.packages:
            if applyFilter(package.ref, filters):
                lines=self.packageWriter.toXML(package, filters, ignore)
                if len(lines)>0:
                    lines=self.indent(lines,2)
                    result+='\n'.join(lines)+'\n'
            ws.unhandledWriter = ws.unhandledWriter.union(package.unhandledWriter)
        lines=self.endFile()
        return result+'\n'.join(lines)+'\n'

    def toCode(self, ws, filters=None, ignore=None, head=None, tail=None, isModule=False, isTemplate=False, indent=3):
        localvars = collections.OrderedDict()
        localvars['ws']=ws
        indentStr=indent*' '
        if isModule == False:
            #head
            if head is None:
                lines=['import autosar']
                if not isTemplate:
                    lines.append('ws=autosar.workspace()')
                result='\n'.join(lines)+'\n\n'
            else:
                if isinstance(head,list):
                    head = '\n'.join(head)
                assert(isinstance(head,str))
                result = head+'\n\n'


            #body
            for package in ws.packages:
                if applyFilter(package.ref, filters):
                    lines=self.packageWriter.toCode(package, filters, ignore, localvars, isTemplate)
                    if len(lines)>0:
                        result+='\n'.join(lines)+'\n'
            #tail
            if not isTemplate:
                if tail is None:
                    result+='\n'+'print(ws.toXML())\n'
                else:
                    if isinstance(tail,list):
                        tail = '\n'.join(tail)
                    assert(isinstance(tail,str))
                    result+='\n'+tail
            return result
        else:
            if head is None:
                head=[
                   ['import autosar'],
                   ['ws = autosar.workspace()']
                ]
            if len(head)!=2:
                raise ValueError('when module=True then head must have exactly two elements (list of lists)')
            if isinstance(head[0], collections.abc.Iterable):
                head[0] = '\n'.join(head[0])
            assert(isinstance(head[0],str))
            result = head[0]+'\n\n'
            #body
            result+='def apply(ws):\n'
            for package in ws.packages:
                if applyFilter(package.ref, filters):
                    lines=self.packageWriter.toCode(package, filters, ignore, localvars, isTemplate)
                    if len(lines)>0:
                        lines=[indentStr+x for x in lines]
                        result+='\n'.join(lines)+'\n'

            #tail
            result+="\nif __name__=='__main__':\n"
            if isinstance(head[1], collections.abc.Iterable):
                head[1] = '\n'.join([indentStr+x for x in head[1]])
            else:
                head[1] = '\n'.join([indentStr+x for x in head[1].split('\n')])
            assert(isinstance(head[1],str))
            result+=head[1]+'\n'
            result+=indentStr+'apply(ws)\n'
            if not isTemplate:
                if tail is None:
                    result+=indentStr+'print(ws.toXML())\n'
                else:
                    if isinstance(tail,list):
                        tail = '\n'.join([indentStr+x for x in tail])
                    else:
                        tail = '\n'.join([indentStr+x for x in tail.split('\n')])
                    assert(isinstance(tail,str))
                    result+=tail+'\n'
            return result

    def saveCode(self, ws, fp, filters=None, ignore=None, head=None, tail=None, isModule=False, isTemplate=False):
        fp.write(self.toCode(ws, filters, ignore, head, tail, isModule, isTemplate))



from autosar.writer.writer_base import ElementWriter
import autosar.behavior

class XMLModeWriter(ElementWriter):
    def __init__(self, version, patch):
        super().__init__(version, patch)

    def getSupportedXML(self):
        return 'ModeDeclarationGroup'

    def getSupportedCode(self):
        return []

    def writeElementXML(self, elem):
        if type(elem).__name__ == 'ModeDeclarationGroup':
            return self.writeModeDeclarationGroupXML(elem)
        else:
            return None

    def writeElementCode(self, elem, localvars):
        raise NotImplementedError('writeElementCode')

    def writeModeDeclarationGroupXML(self, modeDeclGroup):
        assert(isinstance(modeDeclGroup,autosar.mode.ModeDeclarationGroup))
        lines=[]
        ws = modeDeclGroup.rootWS()
        assert(ws is not None)

        lines.append('<%s>'%modeDeclGroup.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%modeDeclGroup.name,1))
        if modeDeclGroup.category is not None:
            lines.append(self.indent('<CATEGORY>%s</CATEGORY>'%modeDeclGroup.category,1))
        if modeDeclGroup.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(modeDeclGroup.adminData),1))
        if modeDeclGroup.initialModeRef is not None:
            modeElem = ws.find(modeDeclGroup.initialModeRef)
            if (modeElem is None):
                raise ValueError("invalid mode reference: '%s'"%modeDeclGroup.typeRef)
            else:
                lines.append(self.indent('<INITIAL-MODE-REF DEST="%s">%s</INITIAL-MODE-REF>'%(modeElem.tag(self.version),modeElem.ref),1))
        if len(modeDeclGroup.modeDeclarations)>0:
            lines.append(self.indent('<MODE-DECLARATIONS>',1))
            for elem in modeDeclGroup.modeDeclarations:
                lines.append(self.indent('<%s>'%elem.tag(self.version),2))
                lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,3))
                if elem.value is not None:
                    lines.append(self.indent('<VALUE>{:d}</VALUE>'.format(elem.value),3))
                lines.append(self.indent('</%s>'%elem.tag(self.version),2))
            lines.append(self.indent('</MODE-DECLARATIONS>',1))
        lines.append('</%s>'%modeDeclGroup.tag(self.version))
        return lines



from autosar.workspace import Workspace
import autosar.component
import autosar.behavior
import autosar.portinterface
import autosar.datatype
import autosar.constant
import autosar.signal
import autosar.package
import autosar.rte
import autosar.builder
import autosar.port
import autosar.element
from abc import (ABC,abstractmethod)
import autosar.base
import autosar.bsw.com
import autosar.bsw.os
import autosar.util

import ntpath
import os
import xml.etree.ElementTree as ElementTree

def workspace(version=3.0, patch = 2, schema=None, attributes=None, useDefaultWriters=True):
   if schema is None and ( (version == 3.0 and patch == 2) or (version == "3.0.2") ):
      schema = 'autosar_302_ext.xsd'
   return autosar.Workspace(version, patch, schema, None, attributes, useDefaultWriters)




def splitRef(ref):
   return autosar.base.splitRef(ref)

def DataElement(name, typeRef, isQueued=False, softwareAddressMethodRef=None, swCalibrationAccess=None, swImplPolicy = None, parent=None, adminData=None):
   return autosar.element.DataElement(name, typeRef, isQueued, softwareAddressMethodRef, swCalibrationAccess, swImplPolicy, parent, adminData)

def ApplicationError(name, errorCode, parent=None, adminData=None):
   return autosar.portinterface.ApplicationError(name, errorCode, parent, adminData)

def ModeGroup(name, typeRef, parent=None, adminData=None):
   return autosar.mode.ModeGroup(name, typeRef, parent, adminData)

def ParameterDataPrototype(name, typeRef, swAddressMethodRef=None, swCalibrationAccess=None, initValue = None, parent=None, adminData=None):
   return autosar.element.ParameterDataPrototype(name, typeRef, swAddressMethodRef, swCalibrationAccess, initValue, parent, adminData)


#template support
class Template(ABC):

   usageCount = 0 #number of times this template have been applied
   static_ref = ''

   @classmethod
   def ref(cls, ws):
      return cls.static_ref

   @classmethod
   def get(cls, ws):
      ref = cls.ref(ws)
      if ws.find(ref) is None:
         ws.apply(cls)
      return ws.find(ref)

   @classmethod
   @abstractmethod
   def apply(cls, ws, **kwargs):
      """
      Applies this class template to the workspace ws
      """



import autosar.component
import autosar.rte.base
import autosar.bsw.com
from collections import namedtuple
#from autosar.rte.base import PortConnector, PortInstance

class Config:
   """
   RTE Configuration
   """
   def __init__(self, partition):
      self.services=[]
      self.partition = partition
      self.data_elements=[]      
      
   
      
      
      



from autosar.element import (Element, DataElement)
import collections
import autosar.base


class ModeGroup(Element):
    """
    Implements <MODE-GROUP> (AUTOSAR4)
    Implements <MODE-DECLARATION-GROUP-PROTOTYPE> (AUTOSAR3)

    A ModeGroup inside a ModeSwitchInterface is what a DataElement is to a SenderReceiverInterface.
    """
    def __init__(self, name, typeRef, parent=None, adminData=None):
        super().__init__(name, parent, adminData)
        self.typeRef=typeRef

    def tag(self,version=None):
        if version>=4.0:
            return "MODE-GROUP"
        else:
            return "MODE-DECLARATION-GROUP-PROTOTYPE"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.name == other.name and self.adminData == other.adminData and self.typeRef == other.typeRef: return True
        return False

    def __ne__(self, other):
        return not (self == other)

class ModeDeclarationGroup(Element):
    """
    Implements <MODE-DECLARATION-GROUP> (AUTOSAR 4)

    Objects created from this class is expected to be placed inside a package with the role name "ModeDclrGroup".

    Attributes:
    modeDeclarations: A list of ModeDeclaration objects
    initialModeRef: Initial mode value

    """
    def tag(self, version=None): return "MODE-DECLARATION-GROUP"

    def __init__(self, name, initialModeRef=None, modeDeclarations=None, category=None, parent=None, adminData=None):
        super().__init__(name, parent, adminData, category)
        self.initialModeRef = initialModeRef
        if modeDeclarations is None:
            self.modeDeclarations = []
        else:
            self.modeDeclarations = list(modeDeclarations)
        self.category=category

    def find(self,ref):
        ref = ref.partition('/')
        name = ref[0]
        for elem in self.modeDeclarations:
            if elem.name==name:
                return elem
        return None

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.name == other.name and self.initialModeRef == other.initialModeRef and \
            len(self.modeDeclarations) == len(other.modeDeclarations) and self.adminData == other.adminData:
                for i,left in enumerate(self.modeDeclarations):
                    right = other.modeDeclarations[i]
                    if left != right:
                        return False
                return True
        return False

    def __ne__(self, other):
        return not (self == other)

class ModeDeclaration(Element):
    """
    Implements <MODE-DECLARATION> (AUTOSAR4)
    """
    def tag(self,version=None): return "MODE-DECLARATION"

    def __init__(self, name, value = None, parent=None, adminData=None):
        super().__init__(name, parent, adminData)
        self.value = int(value) if value is not None else None

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.name == other.name:
                if self.value is None and other.value is None:
                    return True
                elif (self.value is not None) and (other.value is not None):
                    if self.value == other.value:
                        return True
        return False

    def __ne__(self, other): return not (self == other)


import collections
from autosar.element import (Element, DataElement)
import autosar.base
import autosar.mode

class InvalidationPolicy:
    valid_values = ['DONT-INVALIDATE', 'EXTERNAL-REPLACEMENT', 'KEEP', 'REPLACE']

    def tag(self, version): # pylint: disable=unused-argument
        return 'INVALIDATION-POLICY'

    def __init__(self, dataElementRef, handleInvalid):
        self.dataElementRef = dataElementRef
        self.handleInvalid = handleInvalid

    @property
    def handleInvalid(self):
        return self._handleInvalid

    @handleInvalid.setter
    def handleInvalid(self, value):
        if value not in InvalidationPolicy.valid_values:
            raise ValueError('invalid value: %s'%value)
        self._handleInvalid = value

class PortInterface(Element):
    def __init__(self, name, isService=False, serviceKind = None, parent=None, adminData=None):
        super().__init__(name, parent, adminData)
        self.isService=bool(isService)
        self.serviceKind = serviceKind

    def __getitem__(self,key):
        if isinstance(key,str):
            return self.find(key)
        else:
            raise ValueError('expected string')

    def find(self, ref):
        raise NotImplementedError(type(self))

class SenderReceiverInterface(PortInterface):

    def tag(self, version = None): #pylint: disable=unused-argument
        return 'SENDER-RECEIVER-INTERFACE'

    def __init__(self, name, isService=False, serviceKind = None, parent=None, adminData=None):
        super().__init__(name, isService, serviceKind, parent, adminData)
        self.dataElements=[]
        self.modeGroups=[] #AUTOSAR3 only
        self.invalidationPolicies=[] #AUTOSAR4 only

    def __iter__(self):
        return iter(self.dataElements)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.name == other.name and self.isService == other.isService and \
            self.adminData == other.adminData and len(self.dataElements) == len(other.dataElements):
                if (self.modeGroups is not None) and (other.modeGroups is not None) and len(self.modeGroups) == len(other.modeGroups):
                    for i,elem in enumerate(self.modeGroups):
                        if elem != other.modeGroups[i]:
                            return False
                    return True
                for i,elem in enumerate(self.dataElements):
                    if elem != other.dataElements[i]:
                        return False
        return False

    def __ne__(self, other):
        return not (self == other)

    def dir(self):
        return [x.name for x in self.dataElements]

    def find(self,ref):
        ref = ref.partition('/')
        name = ref[0]
        for elem in self.dataElements:
            if elem.name==name:
                return elem
        for elem in self.modeGroups:
            if elem.name==name:
                return elem
        return None

    def append(self,elem):
        """
        Adds child element to this port interface
        """
        if isinstance(elem, DataElement):
            self.dataElements.append(elem)
        elif isinstance(elem, autosar.mode.ModeGroup):
            self.modeGroups.append(elem)
        elif isinstance(elem, InvalidationPolicy):
            self.invalidationPolicies.append(elem)
        else:
            raise ValueError("expected elem variable to be of type DataElement")
        elem.parent=self

class ParameterInterface(PortInterface):
    def tag(self,version=None):
        if version>=4.0:
            return 'PARAMETER-INTERFACE'
        else:
            return 'CALPRM-INTERFACE'

    def __init__(self, name, isService=False, serviceKind = None, parent=None, adminData=None):
        super().__init__(name, isService, serviceKind, parent, adminData)
        self.parameters=[]

    def find(self,ref):
        ref = ref.partition('/')
        name = ref[0]
        for elem in self.parameters:
            if elem.name==name:
                return elem

    def append(self,elem):
        """
        adds elem to the self.parameters list and sets elem.parent to self (the port interface)
        """
        if not isinstance(elem, autosar.element.ParameterDataPrototype):
            raise ValueError("Expected elem variable to be of type ParameterDataPrototype")
        self.parameters.append(elem)
        elem.parent=self

class ClientServerInterface(PortInterface):
    def __init__(self, name, isService=False, serviceKind = None, parent=None, adminData=None):
        super().__init__(name, isService, serviceKind, parent, adminData)
        self.operations=[]
        self.applicationErrors=[]

    def tag(self, version = None): #pylint: disable=unused-argument
        return 'CLIENT-SERVER-INTERFACE'

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.name == other.name and self.adminData == other.adminData and len(self.operations) == len(other.operations) and \
            len(self.applicationErrors) == len(other.applicationErrors):
                for i,operation in enumerate(self.operations):
                    if operation != other.operations[i]: return False
                for i,applicationError in enumerate(self.applicationErrors):
                    if applicationError != other.applicationErrors[i]: return False
                return True
        return False

    def __ne__(self, other):
        return not (self == other)

    def find(self,ref):
        ref = ref.partition('/')
        name = ref[0]
        for elem in self.operations:
            if elem.name==name:
                return elem
        for elem in self.applicationErrors:
            if elem.name==name:
                return elem
        return None

    def append(self,elem):
        """
        Adds elem to the self.operations or self.applicationErrors lists depending on type
        """
        if isinstance(elem, Operation):
            self.operations.append(elem)
        elif isinstance(elem, ApplicationError):
            self.applicationErrors.append(elem)
        else:
            raise ValueError("invalid type: %s"%(str(type(elem))))
        elem.parent=self

class Operation(Element):
    def tag(self,version=None):
        return 'CLIENT-SERVER-OPERATION' if version >=4.0 else 'OPERATION-PROTOTYPE'

    def __init__(self,name,parent=None):
        super().__init__(name,parent)
        self.arguments=[]
        self.errorRefs=[]


    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.name == other.name and self.adminData == other.adminData and len(self.arguments) == len(other.arguments) and \
            (len(self.errorRefs) == len(other.errorRefs)):
                for i,argument in enumerate(self.arguments):
                    if argument != other.arguments[i]: return False
                for i,errorRef in enumerate(self.errorRefs):
                    if errorRef != other.errorRefs[i]: return False
                return True
        return False

    def __ne__(self, other):
        return not (self == other)

    def createOutArgument(self, name, typeRef, swCalibrationAccess = None, serverArgumentImplPolicy=None):
        ws = self.rootWS()
        assert(ws is not None)
        dataType = ws.find(typeRef, role='DataType')
        if dataType is None:
            raise ValueError("invalid name or reference: "+typeRef)
        argument=Argument(name, dataType.ref, 'OUT', swCalibrationAccess, serverArgumentImplPolicy, parent=self)
        self.arguments.append(argument)
        return argument

    def createInOutArgument(self, name, typeRef, swCalibrationAccess = None, serverArgumentImplPolicy=None):
        ws = self.rootWS()
        assert(ws is not None)
        dataType = ws.find(typeRef, role='DataType')
        if dataType is None:
            raise ValueError("invalid name or reference: "+typeRef)
        argument=Argument(name, dataType.ref, 'INOUT', swCalibrationAccess, serverArgumentImplPolicy)
        self.arguments.append(argument)
        return argument

    def createInArgument(self, name, typeRef, swCalibrationAccess = None, serverArgumentImplPolicy=None):
        ws = self.rootWS()
        assert(ws is not None)
        dataType = ws.find(typeRef, role='DataType')
        if dataType is None:
            raise ValueError("invalid name or reference: "+typeRef)
        argument=Argument(name, dataType.ref, 'IN', swCalibrationAccess, serverArgumentImplPolicy)
        self.arguments.append(argument)
        return argument

    def append(self, elem):
        """
        Adds elem to the self.arguments or self.errorRefs lists depending on type
        """
        if isinstance(elem, Argument):
            self.arguments.append(elem)
        elif isinstance(elem, str):
            self.errorRefs.append(elem)
        else:
            raise ValueError("invalid type: %s"%(str(type(elem))))
        elem.parent=self


    @property
    def possibleErrors(self):
        return None

    @possibleErrors.setter
    def possibleErrors(self, values):
        if self.parent is None:
            raise ValueError('cannot call this method without valid parent object')
        if isinstance(values, str):
            values=[values]

        if isinstance(values, collections.abc.Iterable):
            del self.errorRefs[:]
            for name in values:
                found=False
                for error in self.parent.applicationErrors:
                    if error.name == name:
                        self.errorRefs.append(error.ref)
                        found=True
                        break
                if found==False:
                    raise ValueError('invalid error name: "%s"'%name)
        else:
            raise ValueError("input argument must be string or iterrable")


class Argument(Element):
    def tag(self,version=None):
        return 'ARGUMENT-DATA-PROTOTYPE' if version>=4.0 else 'ARGUMENT-PROTOTYPE'

    def __init__(self, name, typeRef, direction, swCalibrationAccess = None, serverArgumentImplPolicy=None, parent=None, adminData=None):
        super().__init__(name, parent, adminData)
        self.typeRef=typeRef
        self.direction = direction
        self.swCalibrationAccess = swCalibrationAccess
        self.serverArgumentImplPolicy = serverArgumentImplPolicy

    @property
    def direction(self):
        return self._direction

    @direction.setter
    def direction(self, value):
        if (value != 'IN') and (value != 'OUT') and (value != 'INOUT'):
            raise ValueError('invalid value :%s'%value)
        self._direction=value

    def __eq__(self, other):
        left_ws = self.rootWS()
        right_ws = other.rootWS()
        assert(left_ws is not None)
        assert(right_ws is not None)
        left_type = left_ws.find(self.typeRef)
        if (left_type is None):
            raise autosar.base.InvalidDataTypeRef(self.typeRef)
        right_type = right_ws.find(other.typeRef)
        if (right_type is None):
            raise autosar.base.InvalidDataTypeRef(other.typeRef)
        if self.direction != other.direction: return False
        if (self.swCalibrationAccess is None and other.swCalibrationAccess is not None) or (self.swCalibrationAccess is not None and other.swCalibrationAccess is None):
            return False
        if self.swCalibrationAccess is not None and other.swCalibrationAccess is not None:
            if self.swCalibrationAccess != other.swCalibrationAccess: return False
        if (self.serverArgumentImplPolicy is None and other.serverArgumentImplPolicy is not None) or (self.serverArgumentImplPolicy is not None and other.serverArgumentImplPolicy is None):
            return False
        if self.serverArgumentImplPolicy is not None and other.serverArgumentImplPolicy is not None:
            if self.serverArgumentImplPolicy != other.serverArgumentImplPolicy: return False
        return left_type == right_type

    def __ne__(self, other):
        return not (self == other)

class ApplicationError(Element):
    def __init__(self, name, errorCode, parent=None, adminData=None):
        super().__init__(name, parent, adminData)
        self.errorCode=int(errorCode)

    def tag(self, version = None):  #pylint: disable=unused-argument
        return 'APPLICATION-ERROR'

    def asdict(self):
        return {'type': self.__class__.__name__, 'name':self.name, 'errorCode':self.errorCode}

class ModeSwitchInterface(PortInterface):
    """
    Implementation of <MODE-SWITCH-INTERFACE> (AUTOSAR 4)
    """

    def tag(self, version = None):  #pylint: disable=unused-argument
        return 'MODE-SWITCH-INTERFACE'

    def __init__(self, name, isService=None, parent=None, adminData=None):
        """
        Arguments:
        name: <SHORT-NAME> (None or str)
        isService: <IS-SERVICE> (None or bool)
        """
        super().__init__(name, isService, parent,adminData)
        self._modeGroup=None

    @property
    def modeGroup(self):
        return self._modeGroup

    @modeGroup.setter
    def modeGroup(self, value):
        if not isinstance(value, autosar.mode.ModeGroup):
            raise ValueError('value must be of ModeGroup type')
        self._modeGroup=value

    def find(self,ref):
        ref = ref.partition('/')
        name = ref[0]
        modeGroup = self.modeGroup
        if modeGroup is not None and modeGroup.name == name:
            return modeGroup
        return None

class NvDataInterface(PortInterface):
    def tag(self,version=None):
        if version>=4.0:
            return 'NV-DATA-INTERFACE'
        else:
            raise ValueError("Autosar 3 is not supported")

    def __init__(self, name, isService=False, serviceKind = None, parent=None, adminData=None):
        super().__init__(name, isService, serviceKind, parent, adminData)
        self.nvDatas=[]

    def find(self,ref):
        ref = ref.partition('/')
        name = ref[0]
        for elem in self.nvDatas:
            if elem.name==name:
                return elem

    def append(self,elem):
        """
        adds elem to the self.nvDatas list and sets elem.parent to self (the port interface)
        """
        if not isinstance(elem, DataElement):
            raise ValueError("expected elem variable to be of type DataElement")
        self.nvDatas.append(elem)
        elem.parent=self



import autosar.constant
import autosar.datatype
import collections
import sys

class ValueBuilder:
    """
    Builds AUTOSAR 4 value specifications from python data
    """
    def __init__(self):
        self.value = None

    def buildFromDataType(self, dataType, rawValue, name = None, ws = None, parent = None):
        assert(dataType is not None)
        if ws is None:
            ws = dataType.rootWS()
            assert(ws is not None)
        return self._createFromDataTypeInternal(ws, name, dataType, rawValue, parent)

    def build(self, label, value):
        return self._createValueInternal(label, value)

    def _createFromDataTypeInternal(self, ws, label, dataType, rawValue, parent = None):
        if isinstance(dataType, (autosar.datatype.ImplementationDataType, autosar.datatype.ApplicationPrimitiveDataType)):
            value = None
            dataConstraint = None
            variantProps = None
            if isinstance(dataType, autosar.datatype.ImplementationDataType):
                variantProps = dataType.variantProps[0]
            if variantProps is not None:
                if variantProps.dataConstraintRef is not None:
                    dataConstraint = ws.find(variantProps.dataConstraintRef, role='DataConstraint')
                    if dataConstraint is None:
                        raise ValueError('{0.name}: Invalid DataConstraint reference: {1.dataConstraintRef}'.format(dataType, variantProps))
                if variantProps.compuMethodRef is not None:
                    compuMethod = ws.find(variantProps.compuMethodRef, role='CompuMethod')
                    if compuMethod is None:
                        raise ValueError('{0.name}: Invalid CompuMethod reference: {1.compuMethodRef}'.format(dataType, variantProps))
                    if compuMethod.category == 'TEXTTABLE':
                        if compuMethod.intToPhys is not None:
                            computation = compuMethod.intToPhys
                        elif compuMethod.physToInt is not None:
                            computation = compuMethod.physToInt
                        else:
                            raise ValueError("CompuMethod {0.name} doesn't have any valid computations")
                        if isinstance(rawValue, int):
                            textValue = computation.getTextValueFromInteger(rawValue)
                            if textValue is None:
                                raise ValueError("Could not find any range in CompuMethod {0.name} matching value {1}".format(compuMethod, rawValue))
                        elif isinstance(rawValue, str):
                            if not computation.verifyTextValue(rawValue):
                                raise ValueError("'{0}' is not a valid value of CompuMethod {1.name}".format(rawValue, compuMethod))
                            textValue = rawValue

                        value = autosar.constant.TextValue(label, str(textValue))
                    else:
                        #TODO: check rawValue here
                        value = autosar.constant.NumericalValue(label, rawValue)
            if value is None:
                if dataType.category == 'VALUE':
                    if isinstance(rawValue, str):
                        value = autosar.constant.TextValue(label, rawValue)
                    else:
                        if dataConstraint is not None:
                            dataConstraint.checkValue(rawValue)
                        value = autosar.constant.NumericalValue(label, rawValue)
                elif dataType.category == 'ARRAY':
                    value = self._createArrayValueFromTypeInternal(ws, label, dataType, rawValue, parent)
                elif dataType.category == 'STRUCTURE':
                    value = self._createRecordValueFromTypeInternal(ws, label, dataType, rawValue, parent)
                elif dataType.category == 'TYPE_REFERENCE':
                    referencedTypeRef = dataType.getTypeReference()
                    referencedType = ws.find(referencedTypeRef, role='DataType')
                    if referencedType is None:
                        raise ValueError('Invalid reference: '+str(referencedTypeRef))
                    value = self._createFromDataTypeInternal(ws, label, referencedType, rawValue, parent)
                else:
                    raise NotImplementedError(dataType.category)
        else:
            raise NotImplementedError(type(dataType))
        return value

    def _createRecordValueFromTypeInternal(self, ws, label, dataType, initValue, parent=None):
        value = autosar.constant.RecordValueAR4(label, dataType.ref, parent=parent)
        if isinstance(initValue, collections.abc.Mapping):
            a = set() #datatype elements
            b = set() #initvalue elements
            for subElem in dataType.subElements:
                a.add(subElem.name)
            for key in initValue.keys():
                b.add(key)
            extra_keys = b-a
            if len(extra_keys) > 0:
                label_str = "" if label is None else "{}: ".format(label)
                raise ValueError('{}Unknown items in initializer: {}'.format(label_str, ', '.join(extra_keys)))


            for elem in dataType.subElements:
                if elem.name in initValue:
                    v = initValue[elem.name]
                    childProps = elem.variantProps[0]
                    if childProps.implementationTypeRef is not None:
                        childTypeRef = childProps.implementationTypeRef
                    else:
                        raise NotImplementedError('could not deduce the type of element "%s"'%(elem.name))
                    childType = ws.find(childTypeRef, role='DataType')
                    if childType is None:
                        raise autosar.base.InvalidDataTypeRef(str(childTypeRef))
                    childValue = self._createFromDataTypeInternal(ws, elem.name, childType, v, value)
                    assert(childValue is not None)
                    value.elements.append(childValue)
                else:
                    name_str = "" if elem.name is None else "{}: ".format(elem.name)
                    raise ValueError('{}Missing initValue field: {}'.format(name_str, elem.name))
        else:
            raise ValueError('initValue must be a dict')
        return value

    def _createArrayValueFromTypeInternal(self, ws, label, dataType, initValue, parent=None):
        value = autosar.constant.ArrayValueAR4(label, dataType.ref, parent=parent)
        typeArrayLength = dataType.subElements[0].arraySize
        if not isinstance(typeArrayLength, int):
            raise ValueError('dataType has no valid array length')
        if isinstance(initValue, collections.abc.Sequence):
            if isinstance(initValue, str):
                initValue = list(initValue)
                if len(initValue)<typeArrayLength:
                    #pad with zeros until length matches
                    initValue += [0]*(typeArrayLength-len(initValue))
                    assert(len(initValue) == typeArrayLength)
            if len(initValue) > typeArrayLength:
                print("%s: Excess array init values detected. Expected length=%d, got %d items"%(label, typeArrayLength, len(initValue)), file=sys.stderr)
            if len(initValue) < typeArrayLength:
                print("%s: Not enough array init values. Expected length=%d, got %d items"%(label, typeArrayLength, len(initValue)), file=sys.stderr)
            i=0
            childTypeRef = dataType.subElements[0].variantProps[0].implementationTypeRef
            childType = ws.find(childTypeRef, role='DataType')
            if childType is None:
                raise autosar.base.InvalidDataTypeRef(str(childTypeRef))
            for v in initValue:
                inner_value = self._createFromDataTypeInternal(ws, None, childType, v, value)
                if inner_value is not None:
                    value.elements.append(inner_value)
                else:
                    raise RuntimeError("Failed to build value for '{}'".format(v))
                i+=1
                if i>=typeArrayLength:
                    break
        return value

    def _createValueInternal(self, label, value):
        if isinstance(value, str):
            return autosar.constant.TextValue(label, value)
        elif isinstance(value, (int, float)):
            return autosar.constant.NumericalValue(label, value)
        elif isinstance(value, collections.abc.Mapping):
            record_value = autosar.constant.RecordValueAR4(label)
            for key in value.keys():
                inner_value = value[key]
                child_value = self._createValueInternal(key, inner_value)
                if child_value is not None:
                    record_value.elements.append(child_value)
                else:
                    raise RuntimeError('Failed to build init-value for "{}"'.format(str(inner_value)))
            return record_value
        elif isinstance(value, collections.abc.Sequence):
            array_value = autosar.constant.ArrayValueAR4(label)
            for inner_value in value:
                child_value = self._createValueInternal(None, inner_value)
                if child_value is not None:
                    array_value.elements.append(child_value)
                else:
                    raise RuntimeError('Failed to build init-value for "{}"'.format(str(inner_value)))
            return array_value
        else:
            raise NotImplementedError(type(value))



"""
Python autosar Ports and ComSpec classes.

Copyright (c) 2019-2026 Conny Gustafsson.
See LICENSE file for additional information.
"""

from autosar.element import Element
import autosar.portinterface
import autosar.constant
import autosar.builder
import copy
import collections
from typing import Optional

sender_receiver_com_spec_arguments_ar4 = {'dataElement', 'initValue', 'initValueRef', 'aliveTimeout', 'queueLength'}
sender_receiver_com_spec_arguments_ar3 = {'dataElement', 'canInvalidate', 'initValueRef', 'aliveTimeout', 'queueLength'}
client_server_com_spec_arguments = {'operation', 'queueLength'}
mode_switch_com_spec_arguments = {'enhancedMode', 'supportAsync', 'queueLength', 'modeSwitchAckTimeout', 'modeGroup'}
parameter_com_spec_arguments = {'parameter', 'initValue'}
nv_data_com_spec_arguments = {'nvData', 'initValue', 'initValueRef', 'ramBlockInitValue', 'ramBlockInitValueRef', 'romBlockInitValue', 'romBlockInitValueRef'}
valid_com_spec_arguments_ar4 = set().union(sender_receiver_com_spec_arguments_ar4, client_server_com_spec_arguments,
    mode_switch_com_spec_arguments, parameter_com_spec_arguments, nv_data_com_spec_arguments)
valid_com_spec_arguments_ar3 = set().union(sender_receiver_com_spec_arguments_ar3, client_server_com_spec_arguments, parameter_com_spec_arguments)

class portOptions:
    """
    Data class that allows the user to store port API options
    to be used later during SwcInternalBehavior creation.
    """
    def __init__(self) -> None:
        self.takeAddress: Optional[bool] = None
        self.indirectAPI: Optional[bool] = None
        self.transformerErrorHandling: Optional[bool] = None

class Port(Element):
    def __init__(self, name, portInterfaceRef, comspec=None, autoCreateComspec = True, options=None, parent=None, adminData=None):
        super().__init__(name, parent, adminData)
        self.comspec=[]
        self.options=portOptions()
        if portInterfaceRef is not None and not isinstance(portInterfaceRef,str):
            raise ValueError('portInterfaceRef needs to be of type None or str')
        self.portInterfaceRef = portInterfaceRef
        ws = self.rootWS()
        assert(ws is not None)
        portInterface=ws.find(portInterfaceRef, role='PortInterface')

        if comspec is None and autoCreateComspec:
            comspec = self._createDefaultComSpecList(portInterface)

        if comspec is not None:
            if isinstance(comspec, ComSpec):
                self.comspec.append(comspec)
            elif isinstance(comspec, collections.abc.Mapping):
                #Typical usage: port interfaces containing a single data element
                comspecObj = self._createComSpecFromDict(ws, portInterface, comspec)
                if comspecObj is None:
                    raise ValueError('Failed to create comspec from comspec data: '+repr(comspec))
                self.comspec.append(comspecObj)
            elif isinstance(comspec, collections.abc.Iterable):
                #Typical usage: port interfaces containing a multiple data elements
                for data in comspec:
                    comspecObj = self._createComSpecFromDict(ws, portInterface, data)
                    if comspecObj is None:
                        raise ValueError('Failed to create comspec from comspec data: '+repr(data))
                    self.comspec.append(comspecObj)
            else:
                raise NotImplementedError("not supported")

        if options is not None:
            if not isinstance(options, dict):
                raise TypeError("optionsDict must be a of type None or dict")
            for key, value in options.items():
                if key == "takeAddress":
                    self.options.takeAddress = bool(value)
                elif key == "indirectAPI":
                    self.options.indirectAPI = bool(value)
                elif key == "transformerErrorHandling":
                    if value is not None:
                        self.options.transformerErrorHandling = bool(value)
                    else:
                        self.options.transformerErrorHandling = value
                else:
                    raise ValueError("Invalid option '{0}'".format(key))

    def _createComSpecFromDict(self, ws, portInterface, comspec):
        """
        Creates ComSpec object based on generic key-value settings from a dictionary.
        """
        self._validate_com_spec_keys_against_port_interface(comspec, portInterface)

        if isinstance(portInterface,autosar.portinterface.SenderReceiverInterface):
            dataElementName = None
            rawInitValue = None
            initValue = None
            initValueRef = None #initValue and initValueRef are mutually exclusive, you cannot have both defined at the same time
            aliveTimeout = 0
            queueLength = None
            canInvalidate = False if isinstance(self,ProvidePort) else None

            if 'dataElement' in comspec: dataElementName=str(comspec['dataElement'])
            if 'initValue' in comspec: rawInitValue=comspec['initValue']
            if 'initValueRef' in comspec: initValueRef=str(comspec['initValueRef'])
            if 'aliveTimeout' in comspec: aliveTimeout=int(comspec['aliveTimeout'])
            if 'queueLength' in comspec: queueLength=int(comspec['queueLength'])
            if 'canInvalidate' in comspec: canInvalidate=bool(comspec['canInvalidate'])
            if (dataElementName is None) and (len(portInterface.dataElements)==1):
                dataElementName=portInterface.dataElements[0].name
            #verify dataElementName
            dataElement=portInterface.find(dataElementName)
            if dataElement is None:
                raise ValueError("Unknown element '%s' of portInterface '%s'"%(dataElementName,portInterface.name))
            #verify compatibility of initValueRef
            if initValueRef is not None:
                initValueTmp = ws.find(initValueRef, role='Constant')
                if initValueTmp is None:
                    raise autosar.base.InvalidInitValueRef(str(initValueRef))
                if isinstance(initValueTmp,autosar.constant.Constant):
                    if ws.version < 4.0:
                        #this is a convenience implementation for the user. For AUTOSAR3, initValueRef needs to point to the value inside the Constant
                        if dataElement.typeRef != initValueTmp.value.typeRef:
                            raise ValueError("constant value has different type from data element, expected '%s', found '%s'"%(dataElement.typeRef,initValue.value.typeRef))
                        initValueRef=initValueTmp.value.ref #correct the reference to the actual value
                    else:
                        initValueRef=initValueTmp.ref
                elif isinstance(initValueTmp,autosar.constant.Value):
                    initValueRef=initValueTmp.ref
                else:
                    raise ValueError("reference is not a Constant or Value object: '%s'"%initValueRef)
            #automatically set default value of queueLength  to 1 in case the dataElement is queued
            if isinstance(self, RequirePort) and dataElement.isQueued and ( (queueLength is None) or queueLength==0):
                queueLength=1
            if rawInitValue is not None:
                valueBuilder = autosar.builder.ValueBuilder()
                if isinstance(rawInitValue, autosar.constant.Value):
                    initValue = rawInitValue
                elif isinstance(rawInitValue, (int, float, str)):
                    dataType = ws.find(dataElement.typeRef, role='DataType')
                    if dataType is None:
                        raise autosar.base.InvalidDataTypeRef(dataElement.typeRef)
                    valueBuilder = autosar.builder.ValueBuilder()
                    initValue = valueBuilder.buildFromDataType(dataType, rawInitValue)
                else:
                    raise ValueError('initValue must be an instance of (autosar.constant.Value, int, float, str)')
            return DataElementComSpec(dataElement.name, initValue, initValueRef, aliveTimeout, queueLength, canInvalidate)
        elif isinstance(portInterface,autosar.portinterface.ClientServerInterface):
            operation = comspec.get('operation', None)
            queueLength = comspec.get('queueLength', 1)
            if operation is not None:
                return OperationComSpec(operation,queueLength)
        elif isinstance(portInterface, autosar.portinterface.ModeSwitchInterface):
            enhancedMode = comspec.get('enhancedMode', ws.profile.modeSwitchEnhancedModeDefault)
            supportAsync = comspec.get('supportAsync', ws.profile.modeSwitchSupportAsyncDefault)
            queueLength = comspec.get('queueLength', None)
            modeSwitchAckTimeout = comspec.get('modeSwitchAckTimeout', None)
            modeGroupName = comspec.get('modeGroup', None)
            if modeGroupName is None:
                modeGroupName = portInterface.modeGroup.name
            if isinstance(self, RequirePort):
                return ModeSwitchComSpec(modeGroupName, enhancedMode, supportAsync, None, None)
            else:
                return ModeSwitchComSpec(modeGroupName, enhancedMode, None, queueLength, modeSwitchAckTimeout)
        elif isinstance(portInterface, autosar.portinterface.ParameterInterface):
            parameterName = comspec.get('parameter', None)
            if (parameterName is None) and (len(portInterface.parameters)==1):
                parameterName = portInterface.parameters[0].name
            initValue = comspec.get('initValue', None)
            return ParameterComSpec(parameterName, initValue)
        elif isinstance(portInterface, autosar.portinterface.NvDataInterface):
            nvDataName = None
            if 'nvData' in comspec: nvDataName=str(comspec['nvData'])
            if (nvDataName is None) and (len(portInterface.nvDatas)==1):
                nvDataName=portInterface.nvDatas[0].name
            #verify nvDataName
            nvData=portInterface.find(nvDataName)
            if nvData is None:
                raise ValueError("Unknown element '%s' of portInterface '%s'"%(nvDataName, portInterface.name))

            if isinstance(self, RequirePort):
                rawInitValue = None
                initValue = None
                initValueRef = None #initValue and initValueRef are mutually exclusive, you cannot have both defined at the same time

                if 'initValue' in comspec: rawInitValue=comspec['initValue']
                if 'initValueRef' in comspec: initValueRef=str(comspec['initValueRef'])
                #verify compatibility of initValueRef
                if initValueRef is not None:
                    initValueTmp = ws.find(initValueRef, role='Constant')
                    if initValueTmp is None:
                        raise autosar.base.InvalidInitValueRef(str(initValueRef))
                    if isinstance(initValueTmp,autosar.constant.Constant):
                        initValueRef=initValueTmp.ref
                    elif isinstance(initValueTmp,autosar.constant.Value):
                        initValueRef=initValueTmp.ref
                    else:
                        raise ValueError("reference is not a Constant or Value object: '%s'"%initValueRef)

                if rawInitValue is not None:
                    if isinstance(rawInitValue, autosar.constant.ValueAR4):
                        initValue = rawInitValue
                    elif isinstance(rawInitValue, (int, float, str)):
                        dataType = ws.find(nvData.typeRef, role='DataType')
                        if dataType is None:
                            raise autosar.base.InvalidDataTypeRef(nvData.typeRef)
                        valueBuilder = autosar.builder.ValueBuilder()
                        initValue = valueBuilder.buildFromDataType(dataType, rawInitValue)
                    else:
                        raise ValueError('initValue must be an instance of (autosar.constant.ValueAR4, int, float, str)')
                return NvRequireComSpec(nvData.name, initValue, initValueRef)
            else:
                # Provide com spec.
                rawRamBlockInitValue = None
                ramBlockInitValue = None
                ramBlockInitValueRef = None #ramBlockInitValue and ramBlockInitValueRef are mutually exclusive, you cannot have both defined at the same time
                rawRomBlockInitValue = None
                romBlockInitValue = None
                romBlockInitValueRef = None #romBlockInitValue and romBlockInitValueRef are mutually exclusive, you cannot have both defined at the same time
                if 'ramBlockInitValue' in comspec: rawRamBlockInitValue=comspec['ramBlockInitValue']
                if 'ramBlockInitValueRef' in comspec: ramBlockInitValueRef=str(comspec['ramBlockInitValueRef'])
                if 'romBlockInitValue' in comspec: rawRomBlockInitValue=comspec['romBlockInitValue']
                if 'romBlockInitValueRef' in comspec: romBlockInitValueRef=str(comspec['romBlockInitValueRef'])

                #verify compatibility of ramBlockInitValueRef
                if ramBlockInitValueRef is not None:
                    initValueTmp = ws.find(ramBlockInitValueRef, role='Constant')
                    if initValueTmp is None:
                        raise autosar.base.InvalidInitValueRef(str(ramBlockInitValueRef))
                    if isinstance(initValueTmp,autosar.constant.Constant):
                        ramBlockInitValueRef=initValueTmp.ref
                    elif isinstance(initValueTmp,autosar.constant.Value):
                        ramBlockInitValueRef=initValueTmp.ref
                    else:
                        raise ValueError("reference is not a Constant or Value object: '%s'"%ramBlockInitValueRef)

                #verify compatibility of romBlockInitValueRef
                if romBlockInitValueRef is not None:
                    initValueTmp = ws.find(romBlockInitValueRef, role='Constant')
                    if initValueTmp is None:
                        raise autosar.base.InvalidInitValueRef(str(romBlockInitValueRef))
                    if isinstance(initValueTmp,autosar.constant.Constant):
                        romBlockInitValueRef=initValueTmp.ref
                    elif isinstance(initValueTmp,autosar.constant.Value):
                        romBlockInitValueRef=initValueTmp.ref
                    else:
                        raise ValueError("reference is not a Constant or Value object: '%s'"%romBlockInitValueRef)

                if rawRamBlockInitValue is not None:
                    if isinstance(rawRamBlockInitValue, autosar.constant.ValueAR4):
                        ramBlockInitValue = rawRamBlockInitValue
                    elif isinstance(rawRamBlockInitValue, (int, float, str)):
                        dataType = ws.find(nvData.typeRef, role='DataType')
                        if dataType is None:
                            raise autosar.base.InvalidDataTypeRef(nvData.typeRef)
                        valueBuilder = autosar.builder.ValueBuilder()
                        ramBlockInitValue = valueBuilder.buildFromDataType(dataType, rawRamBlockInitValue)
                    else:
                        raise ValueError('ramBlockInitValue must be an instance of (autosar.constant.ValueAR4, int, float, str)')

                if rawRomBlockInitValue is not None:
                    if isinstance(rawRomBlockInitValue, autosar.constant.ValueAR4):
                        romBlockInitValue = rawRomBlockInitValue
                    elif isinstance(rawRomBlockInitValue, (int, float, str)):
                        dataType = ws.find(nvData.typeRef, role='DataType')
                        if dataType is None:
                            raise autosar.base.InvalidDataTypeRef(nvData.typeRef)
                        valueBuilder = autosar.builder.ValueBuilder()
                        romBlockInitValue = valueBuilder.buildFromDataType(dataType, rawRomBlockInitValue)
                    else:
                        raise ValueError('romBlockInitValue must be an instance of (autosar.constant.ValueAR4, int, float, str)')
                return NvProvideComSpec(nvData.name, ramBlockInitValue, ramBlockInitValueRef, romBlockInitValue, romBlockInitValueRef)
        else:
            raise NotImplementedError(type(portInterface))
        return None

    def _validate_com_spec_keys_against_port_interface(self, comspec, portInterface):
        if portInterface is None:
            return
        ws = self.rootWS()
        assert(ws is not None)
        if ws.version <= 4.0:
            valid_arguments = valid_com_spec_arguments_ar3
        else:
            valid_arguments = valid_com_spec_arguments_ar4
        for key in comspec.keys():
            if key not in valid_arguments:
                raise ValueError("Unsupported comspec argument '{}'".format(key))

        if isinstance(portInterface,autosar.portinterface.SenderReceiverInterface):
            if ws.version <= 4.0:
                valid_arguments = sender_receiver_com_spec_arguments_ar3
            else:
                valid_arguments = sender_receiver_com_spec_arguments_ar4
        elif isinstance(portInterface,autosar.portinterface.ClientServerInterface):
            valid_arguments = client_server_com_spec_arguments
        elif isinstance(portInterface, autosar.portinterface.ModeSwitchInterface):
            valid_arguments = mode_switch_com_spec_arguments
        elif isinstance(portInterface, autosar.portinterface.ParameterInterface):
            valid_arguments = parameter_com_spec_arguments
        elif isinstance(portInterface, autosar.portinterface.NvDataInterface):
            valid_arguments = nv_data_com_spec_arguments
        else:
            raise RuntimeError("Unsupported port interface:" + str(type(portInterface)))
        for key in comspec.keys():
            if key not in valid_arguments:
                raise ValueError("Comspec argument '{}' doesn't match interface type {}".format(key, type(portInterface)))

    def _createDefaultComSpecList(self, portInterface):
        if portInterface is None:
            return None
        comspecList = []
        if isinstance(portInterface, autosar.portinterface.SenderReceiverInterface):
            for dataElement in portInterface.dataElements:
                comspecList.append({'dataElement': dataElement.name})
        elif isinstance(portInterface,autosar.portinterface.ClientServerInterface):
            for operation in portInterface.operations:
                comspecList.append({'operation': operation.name})
        elif isinstance(portInterface,autosar.portinterface.ParameterInterface):
            for parameter in portInterface.parameters:
                comspecList.append({'parameter': parameter.name})
        elif isinstance(portInterface,autosar.portinterface.NvDataInterface):
            for nvData in portInterface.nvDatas:
                comspecList.append({'nvData': nvData.name})
        elif isinstance(portInterface,autosar.portinterface.ModeSwitchInterface):
            #TODO: Why are we treating ModeSwitchInterface in a different way from the others?
            modeGroupName = portInterface.modeGroup.name
            comspecList.append({'modeGroup': modeGroupName, 'enhancedMode':False, 'supportAsync': False})
        return None if len(comspecList)==0 else comspecList


class RequirePort(Port):
    def tag(self,version=None): return "R-PORT-PROTOTYPE"
    def __init__(self,name , portInterfaceRef=None, comspec=None, autoCreateComSpec = True, optionsDict=None,parent=None):
        if isinstance(name, str):
            #normal constructor
            super().__init__(name, portInterfaceRef, comspec, autoCreateComSpec, optionsDict,parent)
        elif isinstance(name, (RequirePort, ProvidePort)):
            other=name #alias
            #copy constructor
            super().__init__(other.name, other.portInterfaceRef, None, False, other.options, parent)
            self.comspec=copy.deepcopy(other.comspec)
        else:
            raise NotImplementedError(type(name))

    def copy(self):
        """
        returns a copy of itself
        """
        return RequirePort(self)

    def mirror(self):
        """
        returns a mirrored copy of itself
        """
        return ProvidePort(self)


class ProvidePort(Port):
    def tag(self,version=None): return "P-PORT-PROTOTYPE"
    def __init__(self, name, portInterfaceRef = None, comspec = None, autoCreateComSpec = True, optionsDict=None, parent = None):
        if isinstance(name, str):
        #normal constructor
            super().__init__(name, portInterfaceRef, comspec, autoCreateComSpec, optionsDict, parent)
        elif isinstance(name, (RequirePort, ProvidePort)):
            other=name #alias
            #copy constructor
            super().__init__(other.name, other.portInterfaceRef, None, False, other.options,parent)
            self.comspec=copy.deepcopy(other.comspec)
        else:
            raise NotImplementedError(type(name))

    def copy(self):
        """
        returns a copy of itself
        """
        return ProvidePort(self)

    def mirror(self):
        """
        returns a mirrored copy of itself
        """
        return RequirePort(self)

class ComSpec:
    def __init__(self, name=None):
        self.name = name

class OperationComSpec(ComSpec):
    def __init__(self, name=None, queueLength=1):
        super().__init__(name)
        self.queueLength=queueLength

class DataElementComSpec(ComSpec):
    def __init__(self, name=None, initValue=None, initValueRef=None, aliveTimeout=None, queueLength=None, canInvalidate=None, useEndToEndProtection = None):
        super().__init__(name)
        if initValue is not None:
            assert(isinstance(initValue, (autosar.constant.Value, autosar.constant.ValueAR4)))
        self.initValue = initValue
        self.initValueRef = str(initValueRef) if initValueRef is not None else None
        self._aliveTimeout = int(aliveTimeout) if aliveTimeout is not None else None
        self._queueLength = int(queueLength) if queueLength is not None else None
        self.canInvalidate = bool(canInvalidate) if canInvalidate is not None else None
        self.useEndToEndProtection = bool(useEndToEndProtection) if useEndToEndProtection is not None else None

    @property
    def aliveTimeout(self):
        return self._aliveTimeout

    @aliveTimeout.setter
    def aliveTimeout(self,val):
        try:
            self._aliveTimeout = int(val)
        except ValueError:
            self._aliveTimeout = str(val)

    @property
    def queueLength(self):
        return self._queueLength

    @queueLength.setter
    def queueLength(self,val):
        if val is None:
            self._queueLength = None
        else:
            val = int(val)
            if (val > 0):
                self._queueLength = int(val)
            else:
                raise ValueError('queueLength must be None or an int greater than zero')

class ModeSwitchComSpec(ComSpec):
    """
    Implementation of <MODE-SWITCH-SENDER-COM-SPEC> and <MODE-SWITCH-RECEIVER-COM-SPEC>

    Attributes:
    name: Name of the ModeGroup in the associated portInterface (str). This has higher precedence than modeGroupRef.
    enhancedMode: Enables/Disables enhanced mode API (None or bool)
    supportAsync: Enables/Disables support for asynchronous call  (None or bool)
    modeGroupRef: Reference to ModeDeclarationGroup (of the same port interface) (None or str)
    queueLength: Length of call queue on the mode user side (None or int)
    modeSwitchAckTimeout: Timeout (in milliseconds) for acknowledgement of the successful processing of the mode switch request (None or int).
    modeGroupRef: Full mode group reference (None or str). This has lower precendence to name (only used when name is None)
    """
    def __init__(self, name=None, enhancedMode=None, supportAsync=None, queueLength = None, modeSwitchAckTimeout = None, modeGroupRef = None):
        super().__init__(name)
        self.enhancedMode = bool(enhancedMode) if enhancedMode is not None else None
        self.supportAsync = bool(supportAsync) if supportAsync is not None else None
        self._queueLength = int(queueLength) if queueLength is not None else None
        self._modeSwitchAckTimeout = int(modeSwitchAckTimeout) if modeSwitchAckTimeout is not None else None
        self.modeGroupRef = str(modeGroupRef) if modeGroupRef is not None else None

    def tag(self, version, parentPort):
        if isinstance(parentPort, ProvidePort):
            return "MODE-SWITCH-SENDER-COM-SPEC"
        else:
            return "MODE-SWITCH-RECEIVER-COM-SPEC"

    @property
    def queueLength(self):
        return self._queueLength

    @queueLength.setter
    def queueLength(self,val):
        if val is None:
            self._queueLength = None
        else:
            if (val > 0):
                self._queueLength = int(val)
            else:
                raise ValueError('queueLength must be None or an int greater than zero')

    @property
    def modeSwitchAckTimeout(self):
        return self._modeSwitchAckTimeout

    @modeSwitchAckTimeout.setter
    def modeSwitchAckTimeout(self,val):
        if val is None:
            self._modeSwitchAckTimeout = None
        else:
            self._modeSwitchAckTimeout = int(val)

class ParameterComSpec(ComSpec):
    def __init__(self, name, initValue=None):
        super().__init__(name)
        self.initValue = initValue

class NvProvideComSpec(ComSpec):
    """
    Implementation of <NV-PROVIDE-COM-SPEC>

    Attributes:
    name: Name of the NvData in the associated portInterface (str). This has higher precedence than modeGroupRef.
    ramBlockInitValue: Ram block init value.
    ramBlockInitValueRef: Ram block init value reference.
    romBlockInitValue: Rom block init value.
    romBlockInitValueRef: Rom block init value reference.
    variableRef: Full NvData reference (None or str). This has lower precendence to name (only used when name is None)
    """
    def __init__(self, name=None, ramBlockInitValue=None, ramBlockInitValueRef=None, romBlockInitValue=None, romBlockInitValueRef=None, variableRef=None):
        super().__init__(name)
        self.ramBlockInitValue = ramBlockInitValue
        self.ramBlockInitValueRef = str(ramBlockInitValueRef) if ramBlockInitValueRef is not None else None
        self.romBlockInitValue = romBlockInitValue
        self.romBlockInitValueRef = str(romBlockInitValueRef) if romBlockInitValueRef is not None else None
        self.variableRef = str(variableRef) if variableRef is not None else None

    def tag(self, version, parentPort):
            return "NV-PROVIDE-COM-SPEC"

class NvRequireComSpec(ComSpec):
    """
    Implementation of <NV-REQUIRE-COM-SPEC>

    Attributes:
    name: Name of the NvData in the associated portInterface (str). This has higher precedence than modeGroupRef.
    initValue: Init value.
    initValueRef: Ram block init value reference.
    variableRef: Full NvData reference (None or str). This has lower precendence to name (only used when name is None)
    """
    def __init__(self, name=None, initValue=None, initValueRef=None, variableRef=None):
        super().__init__(name)
        self.initValue = initValue
        self.initValueRef = str(initValueRef) if initValueRef is not None else None
        self.variableRef = str(variableRef) if variableRef is not None else None

    def tag(self, version, parentPort):
            return "NV-REQUIRE-COM-SPEC"
